/*
 * $Id: boot.cxx,v 1.14 2002/04/01 13:02:13 cjh Exp $
 *
 * Boot mode downloaders. ROM loader only works with chips containing it!
 *
 * Copyright (c) 1996-2000, Clifford Heath.
 *
 * This program may be used for any commercial or noncommercial purpose
 * without charge, but acknowledgement of authorship and this copyright
 * message must be retained in any copies.
 */
#include	<stdincs.h>

#include	<util.h>
#include	<boot.h>
#include	<db11.h>
#include	<serial.h>

#define	BOOTMAX		1024		/* Max number of bytes to download */
#define	ROMLOADER	0xBF00		/* ROM loader program address */
#define	BOOT_YVAL	0xD000		/* Default Y register value */

#define	BOOT_AND_START "Can't download boot code AND start at non-zero address"

/*
 * Function prototypes:
 */
int	DownLoadROMUsingBoot(FILE* fp, Serial* mcu);
int	DownloadBootImage(FILE* fp, Serial* mcu);
int	DownloadRAM(
		Serial* mcu,		// To this serial port
		char* boot, int s,	// These bytes (which include handshake)
		int chatty		// Verbose?
	);
int	Expect(Serial* mcu, char expect, int nextaddr, int handshake);

/*
 * Download a boot program that
 * sets the Y register to 'yval' and
 * jumps to 'jump'
 */
int
StartAt(Serial* mcu, int jump, int yval)
{
	static	char	start[] = {
			0xFF,			/* Boot handshake byte */
			0x18, 0xCE, 0x00, 0x00,	/* LDY	#0 */
			0x7E, 0x00, 0x00	/* JMP	0 */
		};
#define	YVAL	(start+3)
#define	JUMP	(start+6)

	YVAL[0] = (yval>>8) & 0xFF;
	YVAL[1] = yval & 0xFF;

	JUMP[0] = (jump>>8) & 0xFF;
	JUMP[1] = jump & 0xFF;
	if (not_really || verbose)
		Notify(
			"Reboot jump to 0x%04X with Y=0x%02x\n",
			jump, yval
		);
	if (not_really)
		return TRUE;
	Reset(mcu, "Bootloading jump-start");
	return DownloadRAM(mcu, start, sizeof(start), not_really);
}

/*
 * Download an S19 file to the 68hc11.
 * The boot block segment is read first,
 * then the ROM downloaded (as many segments as necessary),
 * then the boot block is downloaded and run.
 */
int
BootloadS19(FILE* fp, Serial* mcu, uchar bootbyte)
{
	S19	srec;
	char	boot[BOOTMAX+1];
	char	used[BOOTMAX+1];	/* flags to prevent overlapping data */
	int	max = 0;
	int	has_rom = 0;
	int	boot_bytes = 0;
	int	low_address = 0xFFFF;	/* lowest address found */

	/*
	 * Gather code/data for boot RAM (check for ROM stuff too)
	 */
	memset(boot, 0, sizeof(boot));
	memset(used, 0, sizeof(used));
	while (fp && ReadS19(fp, &srec))
	{
		if (srec.address < 0 || srec.address+srec.bytes > BOOTMAX)
		{
			has_rom = 1;
			continue;
		}
		if (srec.bytes && low_address > srec.address)
			low_address = srec.address;
		if (memchr(used+srec.address+1, 1, srec.bytes))
		{
			printf(
				"Overlapping boot memory in S19 at %04lx\n",
				(long)srec.address
			);
			return FALSE;
		}
		memset(used+srec.address+1, 1, srec.bytes);
		memcpy(
			boot+srec.address+1,
			srec.data,
			srec.bytes
		);
		if (max < srec.address+srec.bytes)
			max = srec.address+srec.bytes;
		boot_bytes += srec.bytes;
	}

	/*
	 * Code in ROM must be downloaded first:
	 */
	if (has_rom)
	{
		rewind(fp);
		if (!DownLoadROMUsingBoot(fp, mcu))
			return FALSE;
	}

	/*
	 * If no boot block, download is complete, restart.
	 */
	if (!boot_bytes)
		return TRUE;
	else
		Reset(mcu, "Bootloading RAM");

	/*
	 * Download code/data to boot block:
	 */
	boot[low_address] = bootbyte;
	if (low_address > 0)
	{
		printf(
			"Hmm, boot block starts at 0x%x, must be K series\n",
			low_address
		);
		return DownloadRAM(
				mcu,
				boot+low_address,	// Skip pre-boot data
				max-low_address+1,
				TRUE
			);
	}
	else
		return DownloadRAM(mcu, boot, max+1, TRUE);
}

/*
 * Download an S-record file to ROM, using the ROM burner in the boot ROM.
 * Asks for reset as necessary to jump to new addresses.
 * Any boot block records are skipped, they're handled afterwards.
 */
int
DownLoadROMUsingBoot(FILE* fp, Serial* mcu)
{
	S19		srec;
	unsigned char	expect;
	int		nextaddr = 0xFFFF;
	char*		cp;
	int		vpp_msg = FALSE;
	int		bytes_sent = 0;		/* Bytes done in this block */

	while (ReadS19(fp, &srec))
	{
		if (srec.bytes == 0)
			continue; /* Don't be bothered with empty records */

		/*
		 * Check that data is destined for ROM:
		 */
		if (srec.address >= 0 && srec.address+srec.bytes <= BOOTMAX)
			continue;	/* Skip for now */

		/*
		 * Check that next address is current:
		 */
		if (srec.address != nextaddr)
		{
			nextaddr = srec.address;
			if (!vpp_msg)
			{
				Notify("Apply programming voltage to XIRQ\n");
				vpp_msg = TRUE;
			}
			Reset(mcu, "Loading ROM segment at 0x%04X", nextaddr);
			if (!StartAt(mcu, ROMLOADER, nextaddr))
				Fatal("Failed to start ROM loader");
			bytes_sent = 0;
			expect = 0xFF;
		}

		if (not_really || verbose)
			Notify("Programming %d bytes at 0x%04x\n",
				srec.bytes, nextaddr);
		if (not_really)
		{
			nextaddr += srec.bytes;
			continue;
		}

		/*
		 * Check initial handshake and all echoes while downloading
		 */
		cp = srec.data;
		while (cp < srec.data+srec.bytes)
		{
			if (!Expect(mcu, expect, nextaddr-1, bytes_sent == 0))
				return FALSE;

			/* Send next byte */
			expect = *cp++;
			mcu->write(&expect, 1);
			bytes_sent++;

			nextaddr++;
		}
	}

	/*
	 * Check last echo!
	 */
	if (bytes_sent
	 && !Expect(mcu, expect, nextaddr-1, bytes_sent == 0))
		return FALSE;

	return TRUE;
}

/*
 * The file is a binary image of a boot block.  Download it.
 */
int
DownloadBootImage(FILE* fp, Serial* mcu)
{
	char	boot[BOOTMAX+1];
	int	s;

	if ((s = fread(boot+1, 1, BOOTMAX, fp)) < BOOTMAX
	 && ferror(fp))
		Fatal("read error on boot file");
	boot[0] = 0xFF;		/* Bootstrap handshake byte */

	return DownloadRAM(mcu, boot, s+1, TRUE);
}

/*
 * Download the boot block "boot", of size "s" - including the initial FF!!!
 */
int
DownloadRAM(Serial* mcu, char* boot, int s, int chatty)
{
	unsigned char	back[BOOTMAX+1];
	int		r = 0;
	int		i, j;
	boolean		surely_this_time;

	if (chatty || not_really || verbose)
		Notify("Bootloading %d bytes to boot RAM\n", s-1);

	if (not_really)
		return TRUE;

	/*
	 * All the output must be written in a single write() call, because
	 * if more than four characters of delay occur, the 'E9 will take
	 * off with an incomplete program!  This makes it hard to verify the
	 * echoed data from the download, as you'll see below!
	 */
	mcu->write((unsigned char*)boot, s);

	/*
	 * Cannot tcflush() here or a large download will cause
	 * echoed data to overflow our input buffer.
	 */

	/*
	 * Wait for some output to be echoed, then read it until we would block.
	 * When we have read until we would block, we sleep, and repeat only
	 * if we get some more after sleeping.  We should get about 106 bytes
	 * for each second we sleep.
	 * In any case we must not read more bytes than we downloaded (s-1),
	 * because the program will start producing output too!
	 *
	 * It's preferable if programs have a one-second delay before producing
	 * output at a different baud rate, since we wait up to a second extra
	 * for output at the old baud rate if we miss any of the echoed download
	 */
	sleep(1);	/* Wait for *some* echo */
	surely_this_time = TRUE;
	j = -1;		/* Initial case */
	i = 0;
	while (
		i < s-1	/* We haven't got all we expect */
		&&
		i != j	/* first time, or we got some more for our sleep() */
	)
	{
		j = i;		/* remember how much we had */
		while (i < s-1 && (r = mcu->read(back+i, (s-1)-i, 0)) >= 0)
		{
			if (surely_this_time && r == 0)
				goto pike;	/* Should've had more */
			surely_this_time = FALSE;
			i += r;	/* Got some more! */
		}
		if (r < 0 && errno != EWOULDBLOCK)
			break;	/* sigh, something went wrong, give up! */
		if (i < s-1)
		{
			sleep(1); /* Hmm, we haven't got it all, try again */
			surely_this_time = TRUE;
		}
	}
 pike:
	r = i;

	/*
	 * Check that we got back what we expected:
	 */
	if (r == 0)
		Fatal("No response, check connections and reset");
	else if (s-1 != r || memcmp(boot+1, back, r) != 0)
	{
		int	max = r > s ? r : s;

		i = memdiff(boot+1, (char*)back, r);
		if (i == 255 && r == 256 && back[255] == 0xFF)
		{
			Notify("Last byte echoed as 0xFF, assuming early chip\n");
			mcu->drain();
			return TRUE;
		}
		if (i > 0)
		{
			/*
			 * Some bytes verified ok, choose what message to give
			 */
			Notify(
				"download verify error on boot file:\n"
				"first %d of %d bytes ok, maybe all.\n",
				i, s-1
			);

			if (i == r)
			{
				Notify("I missed the rest, sorry\n");
				i = max; /* Don't dump the missed stuff */
			}
			else
				Notify("differences in the rest are:\n");
		}
		else	/* Perhaps some of the first characters were lost? */
			Notify(
				"download verify error on boot file:\n"
				"Hmm, read %d bytes not matching the %d I wrote:\n",
				r, s-1
			);

		/*
		 * Invoke human pattern matching from here!
		 * This code dumps out the differences in the sent vs received.
		 */
		for (; i < max; i += 20)
		{
			Notify("sent at %d:\t", i);
			for (j = 0; j < 20; j++)
			{
				if (i+j+1 >= s)
					break;
				Notify("%02X ", boot[i+j+1]&0xFF);
			}
			Notify("\n");

			if (i >= r)
				continue;
			Notify("recvd:\t\t");
			for (j = 0; j < 20; j++)
			{
				if (i+j >= r)
					break;
				Notify(
					"%02X%c",
					back[i+j]&0xFF,
					back[i+j] == boot[i+j+1] ? ' ' : '*'
				);
			}
			Notify("\n");
		}
	}
	else
	{
		if (chatty || not_really || verbose)
			Notify("Verified ok\n", s-1);
		mcu->drain();
		return TRUE;
	}
	return force_connect;
}

/*
 * Read one byte from the device, it should be "expect".
 */
int
Expect(Serial* mcu, char expect, int nextaddr, int handshake)
{
	unsigned char	echo;
	int		r;

	/* Receive echoed byte (or handshake byte) */
	r = mcu->read(&echo, 1);
	if (r != 1)
		Fatal("Read error on ROM echo");

	if (echo != expect)
	{
		if (handshake)
		{
			Notify("No ROM handshake, check connections\n");
			return FALSE;
		}

		/* Could check bits and advise erase here */
		Notify(
			"Invalid character programmed in ROM "
			"0x%02X should be 0x%02X at 0x%04X\n",
			echo&0xFF, expect&0xFF, nextaddr
		);
		return FALSE;
	}
	return TRUE;
}
