/*
 * $Id: db11.cxx,v 2.41 2002/04/01 13:02:13 cjh Exp $
 *
 * Download code to a 68hc11 and talk to it in various ways
 *
 * Copyright (c) 1996-2000, Clifford Heath.
 *
 * This program may be used for any commercial or noncommercial purpose
 * without charge, but acknowledgement of authorship and this copyright
 * message must be retained in any copies.
 */
#include	<stdincs.h>

#include	<util.h>
#include	<boot.h>
#include	<midi.h>
#include	<db11.h>
#include	<serial.h>
#include	<comms.h>
#include	<debug.h>
#include	<map.h>
#include	<talker.h>
#include	<symbol.h>

#define	DB11_SCRIPT_PATH	"/usr/local/db11/"

/*
 * External variables, data types:
 */
extern	char*	optarg;
extern	int	optind, opterr, optopt;

/*
 * Function prototypes:
 */
static	void	Usage(int exit_code);
static	void	SuspendMe(int);
static	boolean	ProcessScript(char*);
static	void	ShowModes();
static	void	SetMode(const char* cmd);
boolean		DefineCommand(char**& argv);

/*
 * Global Variables:
 */
char*		progname;		/* db11's name */
int		not_really = FALSE;	/* Just pretend, don't do anything */
int		verbose;		/* Talk a lot */
boolean		modes['z'-'a'+1];	/* Display modes */
Serial*		stdio;			/* I/O object for kbd/screen */
Serial*		mcu;			/* I/O object for MC68HC11 */
int		connectbaud = 1200;	/* Connection baud rate */
int		bootrate = 1200;	/* bootup baud rate */
int		force_connect = FALSE; /* Connect even if boot echo seems bad */

char*		script_dir;		/* dir where current script was found */
const	char*	hex = "0123456789ABCDEF"; /* Hex chars */

int
main(int argc, char** argv)
{
	int	c;
	char*	s19file = 0;	/* s19 file name */
	char*	jump_to = 0;	/* Address to jump to after bootloading */
	char*	cmd[3];
	char**	cmdp;
	boolean	ok = TRUE;

	progname = argv[0];
	cmd[2] = 0;
	while ((c = getopt(argc, argv, "c:fj:m:nr:s:t:?")) != EOF)
	{
		cmdp = cmd;
		cmd[1] = optarg;
		switch (c)
		{
		case 'c':		/* Connect baud rate */
			cmd[0] = "baud";
			Command(cmdp);
			break;

		case 'f':		/* Connect even if boot echo seems bad*/
			cmd[0] = "force";
			cmd[1] = 0;
			Command(cmdp);
			break;

		case 'j':		/* Jump to address given */
			jump_to = optarg;
			break;

		case 'm':
			cmd[0] = "mode";
			Command(cmdp);
			break;

		case 'n':		/* Pretend to load program */
			cmd[0] = "pretend";
			cmd[1] = 0;
			Command(cmdp);
			break;

		case 'r':		/* Bootup baud rate */
			cmd[0] = "bootbaud";
			Command(cmdp);
			break;

		case 's':		/* S19 file to load */
			s19file = optarg;
			break;

		case 't':		/* tty */
			cmd[0] = "tty";
			Command(cmdp);
			break;

		case '?':		/* Help/options */
			Usage(EXIT_SUCCESS);

		default:
			Usage(EXIT_FAILURE);
		}
	}

	/*
	 * Download the appropriate bootstrap
	 */
	if (ok && s19file)			// Bootload an S19 file
	{
		cmd[0] = "boot";
		cmd[1] = s19file;
		cmdp = cmd;
		ok = Command(cmdp);
	}

	if (ok && jump_to)		// just start at given address
	{
		cmd[0] = "jump";
		cmd[1] = jump_to;
		cmdp = cmd;
		ok = Command(cmdp);
	}

	if (!ok && !force_connect)
	{
		delete mcu;
		exit(EXIT_FAILURE);
	}

	if (optind < argc)
		while (ok && optind < argc)
			ok = ProcessScript(argv[optind++]);

	if (ok && mcu)
	{
		cmd[0] = "connect";
		cmd[1] = 0;
		cmdp = cmd;
		ok = Command(cmdp);
	}

	exit(EXIT_SUCCESS);
	/* Not reached */
	return EXIT_SUCCESS;	// Sigh, blasted MSVC...
}

static void
Usage(int exit_code)
{
	fprintf(
		stderr,
		"Usage: %s [options...] [commandfile...]\n"
		"Options"
			"\t-c baud\t\tConnect at baud rate after booting\n"
			"\t-f\t\tConnect even if bootstrap seems to fail\n"
			"\t-j addr\t\tBootstrap by jumping to address\n"
			"\t-m mode\t\tSet connect mode flags:\n"
			"\t\te - local echo\n"
			"\t\th - hex output\n"
			"\t\ti - hex input\n"
			"\t\tm - interpret MIDI protocol\n"
			"\t\tt - talker mode\n"
			"\t-n\t\tDon't really load and connect, just feign it\n"
#if	defined(WINNT)
			"\t-t com\t\tcom port (1-4) where 68hc11 is connected\n"
#else
			"\t-t tty\t\ttty device where 68hc11 is connected\n"
#endif
			"\t-r baud\t\tSet bootup baud rate (1200 default)\n"
			"\t-s s19file\tLoad the S19 file (ROM and/or RAM)\n"
		, progname
	);
	exit(exit_code);
}

#if	defined(__unix)
#if     defined(SIGTSTP)
static void
SuspendMe(int)
{
	stdio->suspend();				// Restore sane modes

	kill(getpid(), SIGSTOP);			// Stop self */
	signal(SIGTSTP, (sighandler_t)SuspendMe);	// Stop next time */

	stdio->resume();				// Resume our modes
}
#endif	/* defined(SIGTSTP) */
#endif

char**	following_commands;

/*
 * Some commands save a copy of all remaining commands from the same line
 * This routine provides the copy.  The act of copying prevents the commands
 * from running now.
 */
void
CommandListSave(char**& argv)
{
	char**		ep;

	// Count words in following commands:
	for (ep = following_commands; ep && *ep && **ep != '#'; ep++)
		;

	argv = new char*[ep-following_commands+1];	// Copy them
	for (ep = following_commands; ep && *ep && **ep != '#'; ep++)
		argv[ep-following_commands] = strdup(*ep);
	argv[ep-following_commands] = 0;	// and terminate the list

	following_commands = 0;		// Don't execute following commands now
}

/*
 * Run a sequence of zero or more commands. Each must advance argv.
 */
boolean
CommandList(
	char** argv,
	boolean (*cmd)(char**& argv)
)
{
	boolean		ok = TRUE;
	char**		sp;
	char**		ep;
	char**		saved_follow = following_commands;
	char*		saved = 0;

	sp = ep = argv;
	while (ok && sp)
	{
		for (ep = sp; *ep && **ep != ';' && **ep != '#'; ep++)
			;		// Find end of command
		following_commands = 0;	// Assume end of command list
		if (*ep)
		{
			saved = *ep;	// Save the command terminator
			*ep = 0;
			if (*saved != '#')	// Not comment
				following_commands = ep+1;
		}

		if (ep > sp)		// Run the command if it's not empty
			ok = (*cmd)(sp);

		if (*sp)
		{			// Warn about unused command words:
			printf("Ignored trailing garbage on command:");
			while (*sp)
				printf(" %s", *sp++);
			printf("\n");
		}

		// Restore previously saved terminator
		if (saved)
		{
			*ep = saved;
			saved = 0;
		}
		sp = following_commands;
	}
	following_commands = saved_follow;
	return ok;
}

boolean
Command(char**& argv)
{
	char*	cmd;
	boolean	ok = TRUE;

#if 0
	printf("Command(");
	for (int i = 0; argv[i]; i++)
		printf("%s%s", i?" ":"", argv[i]);
	printf(")\n");
#endif

	if (!(cmd = *argv))
		return TRUE;	// Null command
	argv++;

	switch (tolower(*cmd))
	{
	case 'b':
		if (strcasecmp("baud", cmd) == 0)
		{
			if (!*argv)
			{
				printf("Usage: baud <rate>\n");
				return TRUE;
			}
			connectbaud = strtol(*argv++, (char**)0, 0);
		set_baud:
			if (mcu)
			{
				mcu->setSpeed(connectbaud);
			}
			return ok;
		}
		else if (strcasecmp("bootbaud", cmd) == 0)
		{
			if (!*argv)
			{
				printf("Usage: bootbaud <rate>\n");
				return TRUE;
			}
			bootrate = strtol(*argv++, (char**)0, 0);
			return TRUE;
		}
		else if (strcasecmp("boot", cmd) == 0)
		{
			/*
			 * Open the S19 file:
			 */
			uchar	bootbyte = 0xFF;
			FILE*	fp = 0;

			if (!*argv)
			{
			bootusage:
				printf("Usage: boot <filename> [ fast | slow | bootbyte ]\n");
				return TRUE;
			}

			if (*argv
			 && (fp = PathOpen(*argv++, "rb", FALSE)) == NULL)
			{
				perror(argv[-1]);
				return FALSE;
			}
			if (*argv)
			{	// Optional third argument to boot
				if (isdigit(**argv))
				{
					bootbyte = strtol(*argv, (char**)0,0);
				}
				else if (strcasecmp(*argv, "f") == 0
				      || strcasecmp(*argv, "fast") == 0)
				{
					bootbyte = 0xF0;
				}
				else if (strcasecmp(*argv, "s") == 0
				      || strcasecmp(*argv, "slow") == 0)
				{
					bootbyte = 0xFF;
				}
				else
					goto bootusage;
				argv++;
			}
			if (!BootloadS19(fp, mcu, bootbyte))
				ok = FALSE;
			fclose(fp);
			if (ok && connectbaud != bootrate)
				goto set_baud;
			return ok;
		}
		break;

	case 'c':
		if (strcasecmp("connect", cmd) == 0)
		{
			/*
			 * Enter connect mode
			 */
			stdio = new Serial();
#if	defined(__unix)
			signal(SIGTSTP, (sighandler_t)SuspendMe);
#endif

			Connect(connectbaud);

			delete stdio;
			return TRUE;
		}
		break;

	case 'd':
		if (strcasecmp("define", cmd) == 0)
			return DefineCommand(argv);
		break;

	case 'f':
		if (strcasecmp("force", cmd) == 0)
		{
			force_connect = !force_connect;
			return TRUE;
		}
		break;

	case '.':
	case 'i':
		if (strcasecmp("include", cmd) == 0
		 || strcasecmp(".", cmd) == 0)
		{
			while (*argv)
				if (!ProcessScript(*argv++))
					break;
			return TRUE;
		}
		break;

	case 'j':
		if (strcasecmp("jump", cmd) == 0)
		{
			int	start_at;

			if (!*argv)
			{
				printf("Usage: jump <address>\n");
				return TRUE;
			}

			start_at = strtol(*argv++, (char**)0, 0);
			StartAt(mcu, start_at, 0);
			return TRUE;
		}
		break;

	case 'm':
		if (strcasecmp("mode", cmd) == 0)
		{
			if (!*argv)
			{
				ShowModes();
				return TRUE;
			}

			while ((cmd = *argv++) != 0)
				SetMode(cmd);

			return TRUE;
		}
		else if (strcasecmp("machine", cmd) == 0)
		{
			MachineSelect(argv);
			return true;
		}
		else if (strcasecmp("map", cmd) == 0)
		{
			MapCommand(argv);
			return true;
		}
		break;

	case 'p':
		if (strcasecmp("pretend", cmd) == 0)
		{
			not_really = !not_really;
			return TRUE;
		}
		break;

	case 's':
		if (strcasecmp("sleep", cmd) == 0)
		{
			int	s = 1;
			if (*argv)
				s = atoi(*argv++);
			if (verbose)
				printf("sleep %d\n", s);
			sleep(s);
			return TRUE;
		}
		break;

	case 't':
		if (strcasecmp("tty", cmd) == 0)
		{
			char*	tty;	/* tty argument for download */

			if (!*argv)
			{
				printf("Usage: tty <ttyname>\n");
				return TRUE;
			}

			if (mcu)	/* Already defined */
				delete mcu;

			tty = *argv++;
#if	defined(__unix)
			if (*tty != '/')
			{
				if (strncmp(tty, "tty", 3) == 0)
					tty = concat("/dev/", tty);
				else
					tty = concat("/dev/tty", tty);
			}

#endif
			mcu = new Serial(tty);
			if (!mcu->isOk())
			{
				Usage(EXIT_FAILURE);
				/* No return */
			}

			/*
			 * Set the mcu into raw mode at the bootup baud rate:
			 */
			mcu->disableSpecialInput();
			mcu->disableSpecialOutput();
			mcu->setSpeed(bootrate);
			return TRUE;
		}
		break;

	default:
		break;
	}
	printf("Unrecognised command %s", cmd);
	while (*argv)
		printf(" %s", *argv++);	// Skip to end of command
	printf("\n");

	return TRUE;	// Ignore illegal commands
}

static	boolean
ProcessScript(char* script)
{
	FILE*		fp = 0;
	char		buf[256];
	char**		argv = 0;
	boolean		ok = TRUE;
	char*		old_script_dir = script_dir;

	fp = PathOpen(script, "r", TRUE);
	if (!fp)
	{
		perror(script);
		return FALSE;
	}

	while (fgets(buf, sizeof(buf), fp))
	{
		int	argc;
		strcmdfree(argv);
		argv = strcmd(buf, ",=@#;", &argc);

		if (DEBUG_MODE)
		{
			if (!CommandList(argv, DebugCommand))
			{
				ok = FALSE;
				break;
			}
			continue;
		}

		if (!CommandList(argv, Command))
		{
			ok = FALSE;
			break;
		}
	}
	strcmdfree(argv);
	fclose(fp);
	delete script_dir;	/* Restore old script_dir */
	script_dir = old_script_dir;
	return ok;
}

/*
 * Open a file, looking in order for it:
 * - where the current script was found
 * - in the current directory
 * - in $HOME/.db11
 * - in DB11_SCRIPT_PATH (/usr/local/db11).
 */
FILE*
PathOpen(const char* file, const char* mode, boolean is_script)
{
	FILE*		fp = 0;
	char*		home;
	char*		dir;
	char*		name;
	unsigned int	i;
	static	char*	incpaths[] = {
		"",
		"./",
		"~/.db11/",
		DB11_SCRIPT_PATH
	};

	for (i = 0; i < sizeof(incpaths)/sizeof(char*); i++)
	{
		dir = incpaths[i];

		if (dir[0] == '~')
		{
		      	if (!(home = getenv("HOME")))
				continue;
			name = new char[strlen(home)+strlen(dir)+strlen(file)];
			strcpy(name, home);
			strcat(name, dir+1);
			strcat(name, file);
		}
		else
		{
			if (dir[0] == '\0'
			 && !(dir = script_dir))
			 	continue;	/* No existing script */
			name = new char[strlen(dir)+strlen(file)+1];
			strcpy(name, dir);
			strcat(name, file);
		}

		if ((fp = fopen(name, mode)) == NULL)
		{
			delete name;
			continue;
		}
		// printf("Found file %s\n", name);

		if (is_script)
		{
			/*
			 * Remember where we found the current script.
			 * Caller should save and restore old script_dir
			 */
			dir = name;
			name = strrchr(dir, '/');
			if (name)
				*name = '\0';
			else
				dir = strdup("./");
			script_dir = dir;
		}
		break;
	}
	return fp;
}

static void
ShowModes()
{
	printf("Modes:\n");
	printf("\tdebug is o%s\n",
		MODEFLAG('d') ? "n" : "ff"
	);
	printf("\techo is o%s\n",
		MODEFLAG('e') ? "n" : "ff"
	);
	printf("\thex output is o%s\n",
		MODEFLAG('h') ? "n" : "ff"
	);
	printf("\tinputhex is o%s\n",
		MODEFLAG('i') ? "n" : "ff"
	);
	printf("\tmidi is o%s\n",
		MODEFLAG('m') ? "n" : "ff"
	);
	printf("\tquiet is o%s\n",
		MODEFLAG('q') ? "n" : "ff"
	);
}

static void
SetMode(const char* cmd)
{
	switch (*cmd)
	{
	case 'd':
		if (strcasecmp("debug", cmd) == 0
		 || strcasecmp("d", cmd) == 0)
		{
			MODEFLAG('d') = !MODEFLAG('d');
			if (MODEFLAG('d') && mcu)
			{
				mcu->setSpeed(connectbaud);
				static boolean	synchronised = FALSE;
				if (!synchronised)
				{
					/* Establish communications */
					if (!TalkerSync())
						return;
					synchronised = TRUE;
					printf("Synchronised\n");
				}
			}
			return;
		}
		break;

	case 'e':
		if (strcasecmp("echo", cmd) == 0
		 || strcasecmp("e", cmd) == 0)
		{
			MODEFLAG('e') = !MODEFLAG('e');
			return;
		}
		break;

	case 'h':
		if (strcasecmp("hex", cmd) == 0
		 || strcasecmp("h", cmd) == 0)
		{
			MODEFLAG('h') = !MODEFLAG('h');
			return;
		}
		break;

	case 'i':
		if (strcasecmp("inputhex", cmd) == 0
		 || strcasecmp("i", cmd) == 0)
		{
			MODEFLAG('i') = !MODEFLAG('i');
			return;
		}
		break;

	case 'm':
		if (strcasecmp("midi", cmd) == 0
		 || strcasecmp("m", cmd) == 0)
		{
			MODEFLAG('m') = !MODEFLAG('m');
			return;
		}
		break;

	case 'q':
		if (strcasecmp("quiet", cmd) == 0
		 || strcasecmp("q", cmd) == 0)
		{
			MODEFLAG('q') = !MODEFLAG('q');
			return;
		}
		break;

	case 's':
		if (strcasecmp("slow", cmd) == 0
		 || strcasecmp("s", cmd) == 0)
		{
			MODEFLAG('s') = !MODEFLAG('s');
			return;
		}

	case 'v':
		if (strcasecmp("verbose", cmd) == 0
		 || strcasecmp("v", cmd) == 0)
		{
			MODEFLAG('v') = !MODEFLAG('v');
			verbose = MODEFLAG('v');
			return;
		}
		break;
	}
	printf("Unknown mode %s\n", cmd);
}

/*
 * Parse an address range.
 *
 * Numbers use C syntax, i.e.:
 * 0NN for octal, 0xMM for hex, [1-9]PP for decimal.
 *
 * Supported range formats are:
 *
 * A	means default or unspecified number of bytes from A
 * A,B	means B bytes starting from A, or (A .. A+B-1)
 * A-B	means (a .. b)
 * ,B	means B bytes from current address
 *
 * Extraneous white space is ignored.
 */
void
GetAddressRange(char**& argv, long* start, long* count)
{
	char*	text;
	char*	cp;
	char	divider;
	long	l;

	if (!(text = *argv))
		return;
	l = strtol(text, &cp, 0);
	if (cp != text)
		*start = l;

	if ((divider = *cp) == '-' || divider == ',')
	{
		if (*++cp == '\0')		// any character after divider?
			if (!(cp = *++argv))	// any following words?
				return;		// so give up
	}
	else if ((cp = *++argv) != 0)
		divider = *cp++;	// Next word must start with divider
	else
		return;				// if there is one...

	*count = 0;
	if (divider == ',' || divider == '-')
	{
		if (*cp == '\0')
			cp = *++argv;	// next word
		text = cp;
		l = strtol(text, &cp, 0);
		if (cp != text)
		{
			*count = l;
			if (divider != ',')	// range rather than start,end
				*count -= *start;
		}
		argv++;		// Done with this word
	}
}

boolean
DefineCommand(char**& argv)
{
	char*		name;
	SymState	state;
	ulong		value;
	char*		opval;
	char		op;

/*
			printf("Define:\n");
			for (char** cpp = argv; *cpp != 0; cpp++)
				printf("\t'%s'\n", *cpp);
			printf("says:\n");
*/

	if ((name = *argv) == 0)
	{
define_usage:
		printf("Usage: define SYMBOL [ @ | = ] VALUE\n");
		return TRUE;
	}
	argv++;

	if (!*argv)
	{		// Inquire the value of symbol
		state = SymbolValue(name, value);
		if (state != SymUndef)
		{
			printf("%s=%ld (%s)\n",
				name,
				value,
				state == SymCode
					? "code"
					: "data"
			);
		}
		else
			printf("Symbol %s not defined\n", name);
		return TRUE;
	}

	/*
	 * (re)define a symbol
	 */
	opval = *argv++;			// *argv must be != 0
	if ((op = *opval) != '@' && op != '=')	// Get code/data operator:
		op = '=';			// Assume code symbol
	else if ((opval = *argv) == 0)		// Get value
		goto define_usage;
	else
		argv++;
	// printf("defining %s %c %s\n", name, op, opval);

	value = strtol(opval, (char**)0, 0);
	SymbolDefine(name, value, op == '=' ? SymCode : SymData);
	return TRUE;
}
