/*
 * $Id: disasm.cxx,v 1.7 2002/03/13 10:17:18 cjh Exp $
 *
 * Disassemble HC11 Instructions.
 *
 * Compile with -DTEST to get an executable program,
 * without it to get just the single-instruction routine.
 *
 * Copyright (c) 1996-2000, Clifford Heath.
 *
 * This program may be used for any commercial or noncommercial purpose
 * without charge, but acknowledgement of authorship and this copyright
 * message must be retained in any copies.
 */
#include	<stdincs.h>

#include	<util.h>
#include	<disasm.h>

/*
 * Condition code bit definitions
 */
#define	CCR_S	0x80
#define	CCR_X	0x40
#define	CCR_H	0x20
#define	CCR_I	0x10
#define	CCR_N	0x08
#define	CCR_Z	0x04
#define	CCR_V	0x02
#define	CCR_C	0x01

typedef	enum
{
	OpIllegal,		/* Illegal instruction (.byte) */
	OpInherent,		/* Operands, if any, are inherent */
	OpPage0,		/* Direct page0 (0..255) addressing */
	OpImmediate,		/* Immediate value */
	OpIndexX,		/* Indexed off X */
	OpIndexY,		/* Indexed off Y */
	OpRelative,		/* Relative branch */
	OpExtended,		/* 2-byte memory address */
	OpBitMask,		/* bset/bclr */
	OpBitBranch,		/* brset/brclr */
	OpBitMaskIndexX,	/* bset/bclr */
	OpBitBranchIndexX,	/* brset/brclr */
	OpBitMaskIndexY,	/* bset/bclr */
	OpBitBranchIndexY,	/* brset/brclr */
} InstrFormat;

#define	UNCOND	0x01		/* Unconditional branch */
#define	COND	0x02		/* Conditional branch */
#define	CALL	0x04		/* Unconditional call/interrupt */

typedef struct
{
	char*		instruction;	/* Opcode name */
	short		operand_len;	/* How many operand bytes? */
	InstrFormat	addr_mode;	/* Addressing mode */
	uchar		branch;		/* Is it an unconditional branch etc? */
} Instruction;

static Instruction instructions[] =
{
	{ "test",	0, OpInherent,		0 }, /*0x00*/
	{ "nop",	0, OpInherent,		0 },
	{ "idiv",	0, OpInherent,		0 },
	{ "fdiv",	0, OpInherent,		0 },
	{ "lsrd",	0, OpInherent,		0 },
	{ "lsld",	0, OpInherent,		0 },
	{ "tap",	0, OpInherent,		0 },
	{ "tpa",	0, OpInherent,		0 },
	{ "inx",	0, OpInherent,		0 },
	{ "dex",	0, OpInherent,		0 },
	{ "clv",	0, OpInherent,		0 },
	{ "sev",	0, OpInherent,		0 },
	{ "clc",	0, OpInherent,		0 },
	{ "sec",	0, OpInherent,		0 },
	{ "cli",	0, OpInherent,		0 },
	{ "sei",	0, OpInherent,		0 },
	{ "sba",	0, OpInherent,		0 }, /*0x10*/
	{ "cba",	0, OpInherent,		0 },
	{ "brset",	3, OpBitBranch,		COND },
	{ "brclr",	3, OpBitBranch,		COND },
	{ "bset",	2, OpBitMask,		0 },
	{ "bclr",	2, OpBitMask,		0 },
	{ "tab",	0, OpInherent,		0 },
	{ "tba",	0, OpInherent,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ "daa",	0, OpInherent,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ "aba",	0, OpInherent,		0 },
	{ "bset",	2, OpBitMaskIndexX,	0 },
	{ "bclr",	2, OpBitMaskIndexX,	0 },
	{ "brset",	3, OpBitBranchIndexX,	COND },
	{ "brclr",	3, OpBitBranchIndexX,	COND },
	{ "bra",	1, OpRelative,		UNCOND }, /*0x20*/
	{ "brn",	1, OpRelative,		0 },
	{ "bhi",	1, OpRelative,		COND },
	{ "bls",	1, OpRelative,		COND },
	{ "bcc",	1, OpRelative,		COND },
	{ "blo",	1, OpRelative,		COND },
	{ "bne",	1, OpRelative,		COND },
	{ "beq",	1, OpRelative,		COND },
	{ "bvc",	1, OpRelative,		COND },
	{ "bvs",	1, OpRelative,		COND },
	{ "bpl",	1, OpRelative,		COND },
	{ "bmi",	1, OpRelative,		COND },
	{ "bge",	1, OpRelative,		COND },
	{ "blt",	1, OpRelative,		COND },
	{ "bgt",	1, OpRelative,		COND },
	{ "ble",	1, OpRelative,		COND },
	{ "tsx",	0, OpInherent,		0 }, /*0x30*/
	{ "ins",	0, OpInherent,		0 },
	{ "pula",	0, OpInherent,		0 },
	{ "pulb",	0, OpInherent,		0 },
	{ "des",	0, OpInherent,		0 },
	{ "txs",	0, OpInherent,		0 },
	{ "psha",	0, OpInherent,		0 },
	{ "pshb",	0, OpInherent,		0 },
	{ "pulx",	0, OpInherent,		0 },
	{ "rts",	0, OpInherent,		UNCOND },
	{ "abx",	0, OpInherent,		0 },
	{ "rti",	0, OpInherent,		UNCOND },
	{ "pshx",	0, OpInherent,		0 },
	{ "mul",	0, OpInherent,		0 },
	{ "wai",	0, OpInherent,		0 },
	{ "swi",	0, OpInherent,		CALL },
	{ "nega",	0, OpInherent,		0 }, /*0x40*/
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ "coma",	0, OpInherent,		0 },
	{ "lsra",	0, OpInherent,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ "rora",	0, OpInherent,		0 },
	{ "asra",	0, OpInherent,		0 },
	{ "lsla",	0, OpInherent,		0 },
	{ "rola",	0, OpInherent,		0 },
	{ "deca",	0, OpInherent,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ "inca",	0, OpInherent,		0 },
	{ "tsta",	0, OpInherent,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ "clra",	0, OpInherent,		0 },
	{ "negb",	0, OpInherent,		0 }, /*0x50*/
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ "comb",	0, OpInherent,		0 },
	{ "lsrb",	0, OpInherent,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ "rorb",	0, OpInherent,		0 },
	{ "aslb",	0, OpInherent,		0 },
	{ "lslb",	0, OpInherent,		0 },
	{ "rolb",	0, OpInherent,		0 },
	{ "decb",	0, OpInherent,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ "incb",	0, OpInherent,		0 },
	{ "tstb",	0, OpInherent,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ "clrb",	0, OpInherent,		0 },
	{ "neg",	1, OpIndexX,		0 }, /*0x60*/
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ "com",	1, OpIndexX,		0 },
	{ "lsr",	1, OpIndexX,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ "ror",	1, OpIndexX,		0 },
	{ "asr",	1, OpIndexX,		0 },
	{ "asl",	1, OpIndexX,		0 },
	{ "rol",	1, OpIndexX,		0 },
	{ "dec",	1, OpIndexX,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ "inc",	1, OpIndexX,		0 },
	{ "tst",	1, OpIndexX,		0 },
	{ "jmp",	1, OpIndexX,		UNCOND },
	{ "clr",	1, OpIndexX,		0 },
	{ "neg",	2, OpExtended,		0 }, /*0x70*/
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ "com",	2, OpExtended,		0 },
	{ "lsr",	2, OpExtended,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ "ror",	2, OpExtended,		0 },
	{ "asr",	2, OpExtended,		0 },
	{ "lsl",	2, OpExtended,		0 },
	{ "rol",	2, OpExtended,		0 },
	{ "dec",	2, OpExtended,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ "inc",	2, OpExtended,		0 },
	{ "tst",	2, OpExtended,		0 },
	{ "jmp",	2, OpExtended,		UNCOND },
	{ "clr",	2, OpExtended,		0 },
	{ "suba",	1, OpImmediate,		0 }, /*0x80*/
	{ "cmpa",	1, OpImmediate,		0 },
	{ "sbca",	1, OpImmediate,		0 },
	{ "subd",	2, OpImmediate,		0 },
	{ "anda",	1, OpImmediate,		0 },
	{ "bita",	1, OpImmediate,		0 },
	{ "ldaa",	1, OpImmediate,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ "eora",	1, OpImmediate,		0 },
	{ "adca",	1, OpImmediate,		0 },
	{ "oraa",	1, OpImmediate,		0 },
	{ "adda",	1, OpImmediate,		0 },
	{ "cpx",	2, OpImmediate,		0 },
	{ "bsr",	1, OpRelative,		CALL },
	{ "lds",	2, OpImmediate,		0 },
	{ "xgdx",	0, OpInherent,		0 },
	{ "suba",	1, OpPage0,		0 }, /*0x90*/
	{ "cmpa",	1, OpPage0,		0 },
	{ "sbca",	1, OpPage0,		0 },
	{ "subd",	1, OpPage0,		0 },
	{ "anda",	1, OpPage0,		0 },
	{ "bita",	1, OpPage0,		0 },
	{ "ldaa",	1, OpPage0,		0 },
	{ "staa",	1, OpPage0,		0 },
	{ "eora",	1, OpPage0,		0 },
	{ "adca",	1, OpPage0,		0 },
	{ "oraa",	1, OpPage0,		0 },
	{ "adda",	1, OpPage0,		0 },
	{ "cpx",	1, OpPage0,		0 },
	{ "jsr",	1, OpPage0,		CALL },
	{ "lds",	1, OpPage0,		0 },
	{ "sts",	1, OpPage0,		0 },
	{ "suba",	1, OpIndexX,		0 }, /*0xa0*/
	{ "cmpa",	1, OpIndexX,		0 },
	{ "sbca",	1, OpIndexX,		0 },
	{ "subd",	1, OpIndexX,		0 },
	{ "anda",	1, OpIndexX,		0 },
	{ "bita",	1, OpIndexX,		0 },
	{ "ldaa",	1, OpIndexX,		0 },
	{ "staa",	1, OpIndexX,		0 },
	{ "eora",	1, OpIndexX,		0 },
	{ "adca",	1, OpIndexX,		0 },
	{ "oraa",	1, OpIndexX,		0 },
	{ "adda",	1, OpIndexX,		0 },
	{ "cpx",	1, OpIndexX,		0 },
	{ "jsr",	1, OpIndexX,		CALL },
	{ "lds",	1, OpIndexX,		0 },
	{ "sts",	1, OpIndexX,		0 },
	{ "suba",	2, OpExtended,		0 }, /*0xb0*/
	{ "cmpa",	2, OpExtended,		0 },
	{ "sbca",	2, OpExtended,		0 },
	{ "subd",	2, OpExtended,		0 },
	{ "anda",	2, OpExtended,		0 },
	{ "bita",	2, OpExtended,		0 },
	{ "ldaa",	2, OpExtended,		0 },
	{ "staa",	2, OpExtended,		0 },
	{ "eora",	2, OpExtended,		0 },
	{ "adca",	2, OpExtended,		0 },
	{ "oraa",	2, OpExtended,		0 },
	{ "adda",	2, OpExtended,		0 },
	{ "cpx",	2, OpExtended,		0 },
	{ "jsr",	2, OpExtended,		CALL },
	{ "lds",	2, OpExtended,		0 },
	{ "sts",	2, OpExtended,		0 },
	{ "subb",	1, OpImmediate,		0 }, /*0xc0*/
	{ "cmpb",	1, OpImmediate,		0 },
	{ "sbcb",	1, OpImmediate,		0 },
	{ "addd",	2, OpImmediate,		0 },
	{ "andb",	1, OpImmediate,		0 },
	{ "bitb",	1, OpImmediate,		0 },
	{ "ldab",	1, OpImmediate,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ "eorb",	1, OpImmediate,		0 },
	{ "adcb",	1, OpImmediate,		0 },
	{ "orab",	1, OpImmediate,		0 },
	{ "addb",	1, OpImmediate,		0 },
	{ "ldd",	2, OpImmediate,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ "ldx",	2, OpImmediate,		0 },
	{ "stop",	0, OpInherent,		0 },
	{ "subb",	1, OpPage0,		0 }, /*0xd0*/
	{ "cmpb",	1, OpPage0,		0 },
	{ "sbcb",	1, OpPage0,		0 },
	{ "addd",	1, OpPage0,		0 },
	{ "andb",	1, OpPage0,		0 },
	{ "bitb",	1, OpPage0,		0 },
	{ "ldab",	1, OpPage0,		0 },
	{ "stab",	1, OpPage0,		0 },
	{ "eorb",	1, OpPage0,		0 },
	{ "adcb",	1, OpPage0,		0 },
	{ "orab",	1, OpPage0,		0 },
	{ "addb",	1, OpPage0,		0 },
	{ "ldd",	1, OpPage0,		0 },
	{ "std",	1, OpPage0,		0 },
	{ "ldx",	1, OpPage0,		0 },
	{ "stx",	1, OpPage0,		0 },
	{ "subb",	1, OpIndexX,		0 }, /*0xe0*/
	{ "cmpb",	1, OpIndexX,		0 },
	{ "sbcb",	1, OpIndexX,		0 },
	{ "addd",	1, OpIndexX,		0 },
	{ "andb",	1, OpIndexX,		0 },
	{ "bitb",	1, OpIndexX,		0 },
	{ "ldab",	1, OpIndexX,		0 },
	{ "stab",	1, OpIndexX,		0 },
	{ "eorb",	1, OpIndexX,		0 },
	{ "adcb",	1, OpIndexX,		0 },
	{ "orab",	1, OpIndexX,		0 },
	{ "addb",	1, OpIndexX,		0 },
	{ "ldd",	1, OpIndexX,		0 },
	{ "std",	1, OpIndexX,		0 },
	{ "ldx",	1, OpIndexX,		0 },
	{ "stx",	1, OpIndexX,		0 },
	{ "subb",	2, OpExtended,		0 }, /*0xf0*/
	{ "cmpb",	2, OpExtended,		0 },
	{ "sbcb",	2, OpExtended,		0 },
	{ "addd",	2, OpExtended,		0 },
	{ "andb",	2, OpExtended,		0 },
	{ "bitb",	2, OpExtended,		0 },
	{ "ldab",	2, OpExtended,		0 },
	{ "stab",	2, OpExtended,		0 },
	{ "eorb",	2, OpExtended,		0 },
	{ "adcb",	2, OpExtended,		0 },
	{ "orab",	2, OpExtended,		0 },
	{ "addb",	2, OpExtended,		0 },
	{ "ldd",	2, OpExtended,		0 },
	{ "std",	2, OpExtended,		0 },
	{ "ldx",	2, OpExtended,		0 },
	{ "stx",	2, OpExtended,		0 }
};

static Instruction instructions0x18[] =
{
	{ 0,		0, OpIllegal,		0 }, /*0x00*/
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ "iny",	0, OpInherent,		0 },
	{ "dey",	0, OpInherent,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 }, /*0x10*/
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ "bset",	2, OpBitMaskIndexY,	0 },
	{ "bclr",	2, OpBitMaskIndexY,	0 },
	{ "brset",	3, OpBitBranchIndexY,	COND },
	{ "brclr",	3, OpBitBranchIndexY,	COND },
	{ 0,		0, OpIllegal,		0 }, /*0x20*/
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ "tsy",	0, OpInherent,		0 }, /*0x30*/
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ "tys",	0, OpInherent,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ "puly",	0, OpInherent,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ "aby",	0, OpInherent,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ "pshy",	0, OpInherent,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 }, /*0x40*/
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 }, /*0x50*/
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ "neg",	1, OpIndexY,		0 }, /*0x60*/
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ "com",	1, OpIndexY,		0 },
	{ "lsr",	1, OpIndexY,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ "ror",	1, OpIndexY,		0 },
	{ "asr",	1, OpIndexY,		0 },
	{ "lsl",	1, OpIndexY,		0 },
	{ "rol",	1, OpIndexY,		0 },
	{ "dec",	1, OpIndexY,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ "inc",	1, OpIndexY,		0 },
	{ "tst",	1, OpIndexY,		0 },
	{ "jmp",	1, OpIndexY,		UNCOND },
	{ "clr",	1, OpIndexY,		0 },
	{ 0,		0, OpIllegal,		0 }, /*0x70*/
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 }, /*0x80*/
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ "cpy",	2, OpImmediate,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ "xgdy",	0, OpInherent,		0 },
	{ 0,		0, OpIllegal,		0 }, /*0x90*/
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ "cpy",	1, OpPage0,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ "suba",	1, OpIndexY,		0 }, /*0xa0*/
	{ "cmpa",	1, OpIndexY,		0 },
	{ "sbca",	1, OpIndexY,		0 },
	{ "subd",	1, OpIndexY,		0 },
	{ "anda",	1, OpIndexY,		0 },
	{ "bita",	1, OpIndexY,		0 },
	{ "ldaa",	1, OpIndexY,		0 },
	{ "staa",	1, OpIndexY,		0 },
	{ "eora",	1, OpIndexY,		0 },
	{ "adca",	1, OpIndexY,		0 },
	{ "oraa",	1, OpIndexY,		0 },
	{ "adda",	1, OpIndexY,		0 },
	{ "cpy",	1, OpIndexY,		0 },
	{ "jsr",	1, OpIndexY,		CALL },
	{ "lds",	1, OpIndexY,		0 },
	{ "sts",	1, OpIndexY,		0 },
	{ 0,		0, OpIllegal,		0 }, /*0xb0*/
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ "cpy",	2, OpExtended,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 }, /*0xc0*/
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ "ldy",	2, OpImmediate,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 }, /*0xd0*/
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ "ldy",	1, OpPage0,		0 },
	{ "sty",	1, OpPage0,		0 },
	{ "subb",	1, OpIndexY,		0 }, /*0xe0*/
	{ "cmpb",	1, OpIndexY,		0 },
	{ "sbcb",	1, OpIndexY,		0 },
	{ "addd",	1, OpIndexY,		0 },
	{ "andb",	1, OpIndexY,		0 },
	{ "bitb",	1, OpIndexY,		0 },
	{ "ldab",	1, OpIndexY,		0 },
	{ "stab",	1, OpIndexY,		0 },
	{ "eorb",	1, OpIndexY,		0 },
	{ "adcb",	1, OpIndexY,		0 },
	{ "orab",	1, OpIndexY,		0 },
	{ "addb",	1, OpIndexY,		0 },
	{ "ldd",	1, OpIndexY,		0 },
	{ "std",	1, OpIndexY,		0 },
	{ "ldy",	1, OpIndexY,		0 },
	{ "sty",	1, OpIndexY,		0 },
	{ 0,		0, OpIllegal,		0 }, /*0xf0*/
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ "ldy",	2, OpExtended,		0 },
	{ "sty",	2, OpExtended,		0 }
};

static Instruction instructions0x1A[] =
{
	{ 0,		0, OpIllegal,		0 }, /*0x00*/
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 }, /*0x10*/
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 }, /*0x20*/
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 }, /*0x30*/
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 }, /*0x40*/
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 }, /*0x50*/
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 }, /*0x60*/
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 }, /*0x70*/
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 }, /*0x80*/
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ "cpd",	2, OpImmediate,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 }, /*0x90*/
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ "cpd",	1, OpPage0,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 }, /*0xa0*/
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ "cpd",	1, OpIndexX,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ "cpy",	1, OpIndexX,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 }, /*0xb0*/
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ "cpd",	1, OpExtended,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 }, /*0xc0*/
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 }, /*0xd0*/
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 }, /*0xe0*/
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ "ldy",	1, OpIndexX,		0 },
	{ "sty",	1, OpIndexX,		0 },
	{ 0,		0, OpIllegal,		0 }, /*0xf0*/
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
};

static Instruction instructions0xCD[] = 
{
	{ 0,		0, OpIllegal,		0 }, /*0x00*/
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 }, /*0x10*/
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 }, /*0x20*/
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 }, /*0x30*/
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 }, /*0x40*/
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 }, /*0x50*/
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 }, /*0x60*/
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 }, /*0x70*/
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 }, /*0x80*/
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 }, /*0x90*/
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 }, /*0xa0*/
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ "cpd",	1, OpIndexY,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ "cpx",	1, OpIndexY,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 }, /*0xb0*/
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 }, /*0xc0*/
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 }, /*0xd0*/
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 }, /*0xe0*/
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ "ldx",	1, OpIndexY,		0 },
	{ "stx",	1, OpIndexY,		0 },
	{ 0,		0, OpIllegal,		0 }, /*0xf0*/
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 },
	{ 0,		0, OpIllegal,		0 }
};

static const char*
addrname(
	ulong	addr,
	const char* (*symbol)(ulong addr, ulong* actual, boolean is_data),
	boolean	is_data = FALSE,
	boolean	offset_ok = TRUE
)
{
	static	char	buf[256];
	ulong		symaddr;	// Actual address
	const char*	name;

	if (symbol
	 && (name = symbol(addr, &symaddr, is_data)) != 0)
	{
		if (offset_ok && addr != symaddr)
		{
			snprintf(
				buf, sizeof(buf),
				"%s+0x%lX",
				name, addr-symaddr
			);
			return buf;
		}
		return name;
	}
	if (addr == 0)
		strcpy(buf, "0");
	else
		snprintf(buf, sizeof(buf), "0x%lX", addr);
	return buf;
}

char*
disas1(
	int	(*getbyte)(void* closure), /* Function to get another byte */
	void	(*ungetbyte)(void* closure, char ch), /* to unget one byte */
	void*	closure,		/* Argument to pass to getbyte */
	ulong*	pc_p,			/* Program counter */
	int	flags,			/* Disassembly flags */
	uchar*	last_was_uncond_p,	/* Flag from last call */
	Predict* predict,		/* Registers to predict branch */
	const char* (*symbol)(ulong addr, ulong* actual, boolean is_data)
)
{
	ulong		instr_pc;	/* PC at start of instruction */
	uchar		binary[10];	/* Bytes of whole instruction */
	ulong		bytes;		/* Number of bytes in instruction */
	ulong		opcode;		/* Current opcode value */
	ulong		opcode2;	/* 2nd byte of 2-byte opcode */
	Instruction*	base;		/* Base of disassembly table */
	Instruction*	instr;		/* Entry in disassembly table */
	char*		opname;		/* Name of instruction */
	ulong		i;
	ulong		operand;	/* one or two operand bytes */
	ulong		brsetclr_op = 0;/* First operand byte for brset/brclr */
	boolean		taken;		/* Was branch taken? */
	ushort		location;	/* memory location for branch predict */
	static char	as_line[80];	/* the output line is sprintf'd here */
	char*		as_out = as_line;
	int		column = 0;	/* Column, for formatting */
	char*		name;		/* Symbol name */
	const char*	sym;
	ulong		symaddr;	/* Value of a symbol */

	*as_out = '\0';
	instr_pc = *pc_p;

	/* get the opcode part of the assembly line */
	opcode = getbyte(closure);
	if (opcode == (ulong)EOF)
		return 0;
	(*pc_p)++;

	bytes = 0;
	binary[bytes++] = opcode;

	switch (opcode)
	{
	default:
		base = instructions;
		instr = base+opcode;
		opcode2 = 0;
		break;

	case 0x18:
		opcode2 = getbyte(closure);
		(*pc_p)++;
		binary[bytes++] = opcode2;
		base = instructions0x18;
		instr = base+opcode2;
		break;

	case 0x1A:
		opcode2 = getbyte(closure);
		(*pc_p)++;
		binary[bytes++] = opcode2;
		base = instructions0x1A;
		instr = base+opcode2;
		break;

	case 0xCD:
		opcode2 = getbyte(closure);
		(*pc_p)++;
		binary[bytes++] = opcode2;
		base = instructions0xCD;
		instr = base+opcode2;
		break;
	}

	opname = instr->instruction;

	/*
	 * Get operands.
	 * If 3 bytes, first byte goes in brsetclr_op,
	 * rest gets shifted into "operand".
	 */
	i = instr->operand_len;
	if (i == 3)
	{
		brsetclr_op = getbyte(closure);
		(*pc_p)++;
		i--;
	}
	operand = 0;
	while (i-- > 0)
	{
		binary[bytes++] = getbyte(closure);
		(*pc_p)++;
		operand = (operand << 8) | binary[bytes-1];
	}

	/*
	 * Extra blank line after unconditional branch if this
	 * instruction isn't another (make jump-tables dense).
	 */
	if (last_was_uncond_p
	 && *last_was_uncond_p
	 && (instr->branch&UNCOND) == 0)
	{
		strcpy(as_out, "\n");
		as_out += strlen(as_out);
	}

	/*
	 * Output the address
	 */
	if ((flags & DISA_NOADDR) == 0)
	{
		sprintf(as_out, "0x%.4lX\t", instr_pc);

		column = 8;
		as_out += strlen(as_out);
	}

	/*
	 * Output the binary code (we expect up to five bytes = 10 chars).
	 */
	if ((flags & DISA_NOBINARY) == 0)
	{
		if (!opname && base != instructions)
			bytes--;		/* Illegal 2-byte opcode */
		for (i = 0; i < bytes; i++)
		{
			sprintf(as_out, "%02X", binary[i]);
			as_out += strlen(as_out);
		}
		column += bytes*2;

		while (column < 24)
		{
			*as_out++ = '\t';
			column = (column+8) & ~0x7;	// tab
		}
		*as_out = '\0';
	}

	if (symbol
	 && (sym = symbol(instr_pc, &symaddr, FALSE)) != 0
	 && symaddr == instr_pc)
	{
		sprintf(
			as_out,
			"%s:\t%s",
			sym,
			strlen(sym) < 7 ? "\t" : ""
		);
		column += strlen(sym)+1;
		column = (column+8) & ~0x7;	// tab
		if (strlen(sym) < 7)
			column += 8;		// extra tab
	}
	else
	{
		sprintf(as_out, "\t\t");
		column += 16;
	}
	as_out += strlen(as_out);

	/*
	 * Output the assembly code.
	 */
	if (predict)
	{
		predict->branch_to = *pc_p;
		predict->call = (instr->branch&CALL) != 0;
	}
	switch (instr->addr_mode)
	{
	case OpInherent:
		sprintf(as_out, "%s", opname);

		if (!predict)
			break;
		if (instr->branch&UNCOND)
		{		/* rti, rts */
			if (opcode == 0x39)
				location = predict->sp+1;	/* rts */
			else
				location = predict->sp+8;	/* rti */
			predict->branch_to = predict->getmem(location)<<8
					   | predict->getmem(location+1);
		}
		else if (instr->branch&CALL)
		{		/* swi */
			/*
			 * REVISIT: SWI vector unknown.
			 * SWI vector location depends on whether chip's in
			 * bootstrap or normal mode, and that depends on
			 * MODA bit, which is in the device registers at a
			 * location that depends on CONFIG, which is also
			 * in the device registers... sigh, snookered.
			 *
			 * It doesn't matter for db11, because SWI should
			 * jump to the talker, so we don't want to stop it.
			 */
			// predict->branch_to = ?
		}
		break;

	case OpIndexX:
		sprintf(
			as_out, "%s\t%s,X",
			opname, addrname(operand, symbol, TRUE, FALSE)
		);
		if (predict
		 && instr->branch&(UNCOND|CALL))
		{		// jmp, jsr
			predict->branch_to = predict->x+operand;
		}
		break;

	case OpIndexY:
		sprintf(
			as_out, "%s\t%s,Y",
			opname, addrname(operand, symbol, TRUE, FALSE)
		);
		if (predict
		 && instr->branch&(UNCOND|CALL))
		{		// jmp, jsr
			predict->branch_to = predict->y+operand;
		}
		break;

	case OpRelative:
		sprintf(as_out, "%s\t%s", opname,
			addrname(((*pc_p)+(signed char)operand)&0xFFFF, symbol)
		);
		if (!predict)
			break;
		taken = TRUE;
		switch (opcode)
		{
		case 0x8D:	/* bsr */
		case 0x20:	/* bra */
			break;
		case 0x21:	/* brn */
			taken = FALSE;
			break;
		case 0x22:	/* bhi */
			taken = (predict->ccr&CCR_C) == 0 && (predict->ccr&CCR_Z) == 0;
			break;
		case 0x23:	/* bls */
			taken = (predict->ccr&CCR_C) != 0 || (predict->ccr&CCR_Z) != 0;
			break;
		case 0x24:	/* bcc C==0? (bhs) */
			taken = (predict->ccr&CCR_C) == 0;
			break;
		case 0x25:	/* blo C==1? (bcs) */
			taken = (predict->ccr&CCR_C) != 0;
			break;
		case 0x26:	/* bne Z==0? */
			taken = (predict->ccr&CCR_Z) == 0;
			break;
		case 0x27:	/* beq Z==1? */
			taken = (predict->ccr&CCR_Z) != 0;
			break;
		case 0x28:	/* bvc V==0? */
			taken = (predict->ccr&CCR_V) == 0;
			break;
		case 0x29:	/* bvs V==1? */
			taken = (predict->ccr&CCR_V) != 0;
			break;
		case 0x2A:	/* bpl N==0? */
			taken = (predict->ccr&CCR_N) == 0;
			break;
		case 0x2B:	/* bmi N==1? */
			taken = (predict->ccr&CCR_N) != 0;
			break;
		case 0x2C:	/* bge */
			taken = ((predict->ccr&CCR_N)==0) == ((predict->ccr&CCR_V)==0);
			break;
		case 0x2D:	/* blt */
			taken = ((predict->ccr&CCR_N)==0) != ((predict->ccr&CCR_V)==0);
			break;
		case 0x2E:	/* bgt */
			taken = (predict->ccr&CCR_Z) == 0 && 
				((predict->ccr&CCR_N)==0) == ((predict->ccr&CCR_V)==0);
			break;
		case 0x2F:	/* ble */
			taken = (predict->ccr&CCR_Z) != 0 || 
				((predict->ccr&CCR_N)==0) != ((predict->ccr&CCR_V)==0);
			break;
		}
		predict->branch_to = taken
					? ((*pc_p)+(signed char)operand)&0xFFFF
					: *pc_p;
		break;

	case OpBitMask:		/* bset/bclr */
		name = strdup(addrname(operand>>8, symbol, TRUE, FALSE));
		sprintf(as_out, "%s\t%s,%s",
			opname,
			name, addrname(operand&0xFF, symbol, TRUE, FALSE)
		);
		delete [] name;
		break;

	case OpBitBranch:	/* brset/brclr */
		name = strdup(addrname(operand>>8, symbol, TRUE, FALSE));
		sprintf(as_out, "%s\t0x%lX %s %s",
			opname, brsetclr_op, name,
			addrname(((*pc_p)+(signed char)(operand&0xFF))&0xFFFF, symbol)
		);
		delete [] name;
		if (!predict)
			break;

		location = brsetclr_op;
	brsetclr:
		if (opcode == 0x12)
		{		// brset
			if ((predict->getmem(location) & (operand>>8)) == 0)
				break; 	/* not taken */
		}
		else
		{		// brclr
			if ((~predict->getmem(location) & (operand>>8)) == 0)
				break; 	/* not taken */
		}
		predict->branch_to = ((*pc_p)+(signed char)(operand&0xFF))&0xFFFF;
		break;

	case OpBitMaskIndexX:	/* bset/bclr */
		name = strdup(addrname(operand>>8, symbol, TRUE, FALSE));
		sprintf(as_out, "%s\t%s,X,%s",
			opname, name,
			addrname(operand&0xFF, symbol, TRUE, FALSE));
		delete [] name;
		break;

	case OpBitBranchIndexX:	/* brset/brclr */
		name = strdup(addrname(operand>>8, symbol, TRUE, FALSE));
		sprintf(as_out, "%s\t0x%lX,X %s %s",
			opname, brsetclr_op, name,
			addrname(((*pc_p)+(signed char)(operand&0xFF))&0xFFFF, symbol)
		);
		delete [] name;
		if (!predict)
			break;

		location = predict->x + (operand>>8);
		goto brsetclr;

	case OpBitMaskIndexY:	/* bset/bclr */
		name = strdup(addrname(operand>>8, symbol, TRUE, FALSE));
		sprintf(as_out, "%s\t%s,Y,%s",
			opname, name, addrname(operand&0xFF, symbol, TRUE, FALSE));
		delete [] name;
		break;

	case OpBitBranchIndexY:	/* brset/brclr */
		name = strdup(addrname(operand>>8, symbol, TRUE, FALSE));
		sprintf(as_out, "%s\t0x%lX,Y %s %s",
			opname, brsetclr_op, name,
			addrname(((*pc_p)+(signed char)(operand&0xFF))&0xFFFF, symbol)
		);
		delete [] name;
		if (!predict)
			break;

		location = predict->y + (operand>>8);
		goto brsetclr;

	case OpPage0:
		sprintf(as_out, "%s\t%s", opname,
			addrname(
				operand, symbol,
				(instr->branch&(UNCOND|CALL)) == 0,
				FALSE
			)
		);
		if (predict
		 && instr->branch&(UNCOND|CALL))
		{		// jsr
			predict->branch_to = operand;
		}
		break;

	case OpExtended:
		sprintf(as_out, "%s\t%s", opname,
			addrname(
				operand, symbol,
				(instr->branch&(UNCOND|CALL)) == 0,
				FALSE
			)
		);
		if (predict
		 && instr->branch&(UNCOND|CALL))
		{		// jmp, jsr
			predict->branch_to = operand;
		}
		break;

	case OpImmediate:
		sprintf(as_out, "%s\t#%s",
			opname, addrname(operand, symbol, TRUE, FALSE));
		break;

	case OpIllegal:
		/*
		 * We output one byte here, then if it was a
		 * 2-byte opcode, we disassemble the other byte.
		 */
		sprintf(as_out, ".byte\t0x%lX", opcode);
		break;
	}
	as_out += strlen(as_out);

	if (!opname
	 && base != instructions)
	{
		/* Illegal 2-byte opcode, re-use 2nd byte */
		ungetbyte(closure, opcode2);
		(*pc_p)--;
	}

	/* Extra blank line after unconditional jump instructions */
	if (last_was_uncond_p)
		*last_was_uncond_p = (instr->branch&UNCOND) != 0;
	return as_line;
}

#ifdef	TEST

void
disas(
	int	(*getbyte)(void* closure), /* Function to get another byte */
	void	(*ungetbyte)(void* closure, char ch), /* to unget one byte */
	void*	closure,		/* Argument to pass to getbyte */
	ulong	pc,			/* Program counter */
	ulong	end,			/* end address */
	int	flags			/* Disassembly flags */
)
{
	uchar	extra = FALSE;
	char*	cp;

	if (end == 0)			/* do whole file */
		end = 0xffffffff;

	while (pc <= end
	 && (cp = disas1(getbyte, ungetbyte, closure, &pc, flags, &extra)))
		printf("%s\n", cp);

	putchar('\n');
}

int
getbyte(void* closure)
{
	return getc((FILE*)closure);
}

void
ungetbyte(void* closure, char ch)
{
	ungetc(ch, (FILE*)closure);
}

int
main(int argc, char** argv)
{
	char**		cpp;
	FILE*		fp;
	unsigned long	start = 0, end = 0;
	int		no_address = FALSE;
	int		no_binary = FALSE;
	char*		my_name = argv[0];

	if (argc == 1)
		goto croak2;
	setbuf(stdout, 0);
	for (cpp = &argv[1]; *cpp; ++cpp)
	{
		if (**cpp == '-')
		{
			switch (*(*cpp + 1))
			{
			case 'a':	/* Omit address field from output */
				no_address = TRUE;
				continue;
			case 'b':
				no_binary = TRUE;
				continue;
			case 's':	/* start address */
				if (**(++cpp))
					sscanf(*cpp, "%lx", &start);
				else
					goto croak;
				continue;
			case 'e':	/* end address */
				if (**(++cpp))
					sscanf(*cpp, "%lx", &end);
				else
					goto croak;
				continue;
			default:
			croak:
				fprintf(stderr,
					"%s: Invalid command line.\n",
					my_name
				);
			croak2:
				fprintf(stderr,
					"%s [[-s start] [-e end] [filename]]\n",
					my_name
				);
				exit(1);
			}
		}
		else if ((fp = fopen(*cpp, "rb")) == NULL)
		{
			fprintf(stderr, "%s: Can't open %s.\n", my_name, *cpp);
			continue;
		}

		/* seek to starting offset in file */
		fseek(fp, start, SEEK_SET);

		/*
		 * Print any required heading:
		 */
		if (!no_address)
			printf("Addr\t\t");
		if (!no_binary)
			printf("Instruction");
		if (!no_binary || !no_address)
			printf("\tDisassembled\n\n");

		disas(getbyte, ungetbyte, fp, start, end,
			(no_address ? DISA_NOADDR : 0)
			| (no_binary ? DISA_NOBINARY : 0)
		);

		fclose(fp);
		start = end = 0;
	}
	exit(0);
}
#endif	/* TEST */
