/*
 * $Id: map.cxx,v 1.12 2002/04/14 23:14:32 cjh Exp $
 *
 * Memory maps.
 *
 * Each memory map is named by a "machine name".
 * Only one map is active at a time, though all will be remembered.
 *
 * Each map contains zero or more memory types,
 * and zero or more memory regions (each of a given type)
 * Each region covers a single address range of one memory type.
 *
 * The machine list is managed through the Machine*() functions.
 * The memory type list is managed through the MemType*() functions.
 * The regions list is managed through the Region*() functions.
 * The external API is through the Map*() functions.
 *
 * Still a bit primitive; memory types should be defined using script commands.
 *
 * Copyright (c) 1996-2000, Clifford Heath.
 *
 * This program may be used for any commercial or noncommercial purpose
 * without charge, but acknowledgement of authorship and this copyright
 * message must be retained in any copies.
 */
#include	<stdincs.h>

#include	<ctype.h>

#include	<util.h>
#include	<db11.h>
#include	<map.h>

#define	AllocRound(n)	(((n)+0xF) & ~0xF)
#define	AllocNeeded(was, incr)	(AllocRound(was) != AllocRound((was)+(incr)))

struct Machine
{
	const char*	name;
	MemRegion*	region;
	int		num_region;
	MemType**	mem_type;
	int		num_type;
};

/*
 * Machine table:
 */
static	Machine*	machine;	// Current machine
static	Machine*	machines;	// All machines
static	int		num_machine;	// Number of machines

static	void		MachineInitAll(); // Create initial list of machines

/*
 * Global memory types
 */
static	MemType		ram =
{
	"ram",
	0					// Nothing special to do.
};

static	MemType		eeprom =
{
	"eeprom",
	0x103B, 10, FALSE,			// EPROG
	0x02, 0x03, 0,				// program prep/start/end
	0x06, 0x07, 0,				// erase all prep/start/end
	0x16, 0x17, 0				// erase byte prep/start/end
};

static	MemType		eprom =
{
	"eprom",
	0x103B, 2,  TRUE,			// EPROG, 2ms, ext voltage
	0x20, 0x21, 0,				// program prep/start/end
	0,    0,    0,				// erase all prep/start/end
	0,    0,    0				// erase byte prep/start/end
};

static	MemType		eprom2 =
{
	"eprom2",
	0x1036, 2,  TRUE,			// EPROG2, 2ms, ext voltage
	0x20, 0x21, 0,				// program prep/start/end
	0,    0,    0,				// erase all prep/start/end
	0,    0,    0				// erase byte prep/start/end
};

static	MemType	*GlobalMemType[]
	= { &ram, &eeprom, &eprom, &eprom2 };
#define	GLOBAL_MEM_TYPES	(int)(sizeof(GlobalMemType)/sizeof(MemType*))

/*
 * Find a named machine in the machines table:
 */
static Machine*
MachineFind(const char* machine_name)
{
	int		i;

	for (i = 0; i < num_machine; i++)
		if (strcasecmp(machine_name, machines[i].name) == 0)
			return machines+i;
	return 0;
}

/*
 * Add a new machine to the machines table, making it the current one:
 */
static void
MachineAdd(const char* machine_name)
{
	/*
	 * Resize the machines array:
	 */
	if (AllocNeeded(num_machine, 1))
	{
		machine = new Machine[AllocRound(num_machine+1)];
		if (num_machine)
		{
			memcpy(machine, machines, sizeof(Machine)*num_machine);
			delete machines;
		}
		machines = machine;
	}

	/*
	 * Fill in the new machine entry:
	 */
	machine = machines+num_machine++;
	machine->name = strdup(machine_name);
	machine->region = 0;
	machine->num_region = 0;
	machine->mem_type = 0;
	machine->num_type = 0;
}

/*
 * List the names of machines in the machines table
 */
boolean
MachineShowAll()
{
	int	i;

	MachineInitAll();

	printf("Machine maps are defined for:\n\t");
	for (i = 0; i < num_machine; i++)
		printf("%s%s", i ? " " : "", machines[i].name);
	printf("\n");
	return TRUE;
}

/*
 * Memory types table, one per Machine.
 */
static void
MemTypeAdd(MemType* mem_type)
{
	MachineInitAll();

	/*
	 * Resize the memory types array:
	 */
	if (AllocNeeded(machine->num_type, 1))
	{
		MemType**	new_types;

		new_types = new MemType*[AllocRound(machine->num_type+1)];
		if (machine->num_type)
		{
			memcpy(
				new_types,
				machine->mem_type,
				sizeof(MemType)*machine->num_type
			);
			delete machine->mem_type;
		}
		machine->mem_type = new_types;
	}

	/*
	 * Fill in the new member:
	 */
	machine->mem_type[machine->num_type++] = mem_type;
}

/*
 * Find the named memory type definition in this machine
 * or add it from the global list or memory types.
 */
static MemType*
MemTypeFind(const char* mem_type_name)
{
	int		i;

	for (i = 0; i < machine->num_type; i++)
		if (strcasecmp(machine->mem_type[i]->name, mem_type_name) == 0)
			return machine->mem_type[i];

	for (i = 0; i < GLOBAL_MEM_TYPES; i++)
		if (strcasecmp(GlobalMemType[i]->name, mem_type_name) == 0)
		{
			// Add the memory type from the global list
			MemTypeAdd(GlobalMemType[i]);
			return machine->mem_type[machine->num_type-1];
		}
	return 0;
}

/*
 * Memory regions
 */
static void
RegionShow(MemRegion* region)
{
	printf(
		"\tFrom 0x%04X to 0x%04X is %s\n",
		region->start_addr,
		region->end_addr ? region->end_addr : 0x10000,
		region->mem_type ? region->mem_type->name : "ram"
	);
}

/*
 * Display regions in specified map (or current map, if machine_name is NULL)
 */
boolean
RegionShowAll(char* machine_name)
{
	Machine*	mp = machine;
	int		i;

	MachineInitAll();

	if (machine_name
	 && !(mp = MachineFind(machine_name)))
	{
		printf("Machine %s not defined\n", machine_name);
		return FALSE;
	}

	printf("Map for %s:\n", mp->name);
	for (i = 0; i < mp->num_region; i++)
		RegionShow(mp->region+i);
	return TRUE;
}

/*
 * Add a region in the map for the current machine
 */
static void
RegionAdd(const MemRegion* region)
{
	MachineInitAll();

	/*
	 * Resize the memory regions array:
	 */
	if (AllocNeeded(machine->num_region, 1))
	{
		MemRegion*	new_regions;

		new_regions = new MemRegion[AllocRound(machine->num_region+1)];
		if (machine->num_region)
		{
			memcpy(
				new_regions,
				machine->region,
				sizeof(MemRegion)*machine->num_region
			);
			delete machine->region;
		}
		machine->region = new_regions;
	}

	/*
	 * Fill in the new member:
	 */
	machine->region[machine->num_region++] = *region;
}

/*
 * Find the smallest region containing the address "base",
 * or exact match for start,count if count is specified
 */
static int
RegionFind(unsigned int base, unsigned int count = 0)
{
	MemRegion*	region;
	int		i;
	int		smallest = -1;
	unsigned int	min_size = 0x10000;

	for (
		i = 0, region = machine->region;
		i < machine->num_region;
		i++, region++
	)
	{
		if (base >= region->start_addr
		 && (base < region->end_addr || region->end_addr == 0))
		{
			if (count
			 && region->Count() == count
			 && base == region->start_addr)
				return i;	// Exact match
			if (min_size >= region->Count())
			{
				min_size = region->Count();
				smallest = i;
			}
		}
	}
	return smallest;
}

static void
RegionDelete(int region)
{
	if (region < 0
	 || region >= machine->num_region)
	 	return;	// Error
	memmove(
		machine->region+region,
		machine->region+region+1,
		sizeof(MemRegion)*(machine->num_region-region-1)
	);
	machine->num_region--;
}

/*
 * Command to show name of current machine map, or create a new map.
 */
boolean
MachineSelect(char**& argv)
{
	MachineInitAll();

	if (!*argv)
	{		/* Show name of current machine map */
		printf("Current machine map is %s\n", machine->name);
		return TRUE;
	}

	/*
	 * Set current map
	 */
	Machine*	mp = MachineFind(*argv);
	if (mp)
	{
		machine = mp;
		argv++;
		return TRUE;
	}

	MachineAdd(*argv);
	printf("Adding machine map %s\n", *argv);
	argv++;

#if 0
	/*
	 * Not found, list valid maps
	 */
	printf("Machine map %s unknown, ", *argv);
	MachineShowAll();
	argv++;
#endif
	return TRUE;
}

// Add or modify a map entry
boolean
MapCommand(char**& argv)
{
	MachineInitAll();

	if (!*argv)
	{		/* Show name of current machine map */
		RegionShowAll(0);
		return TRUE;
	}

	/*
	 * Usage: map start end memory-type
	 */
	long		start = 0, count = 0;
	const	char*	mem_type_name = "ram";

	GetAddressRange(argv, &start, &count);

	if (*argv)
		mem_type_name = *argv++;
	else
		printf("Assuming ram\n");

	int	r = RegionFind(start, count);

	if (strcasecmp(mem_type_name, "delete") == 0)
	{
		if (r < 0)
		{
			printf("Map entry not found at 0x%lX\n", start);
			return FALSE;
		}
		RegionDelete(r);
		return TRUE;
	}

	if (count == 0)
	{
		printf("New map entry needs a non-zero size\n");
		return FALSE;
	}

	// Replace an existing region only if the size matches:
	if (r >= 0
	 && start == machine->region[r].start_addr
	 && (ushort)(start+count) == machine->region[r].end_addr)
	{
		printf("Replacing region ");
		RegionShow(machine->region+r);
		RegionDelete(r);
	}

	MemType*	mem_type = MemTypeFind(mem_type_name);

	if (!mem_type)
	{
		printf("Memory type %s unknown\n", mem_type_name);
		return FALSE;
	}

	MemRegion	region;
	region.start_addr = start;
	region.end_addr = start+count;
	region.mem_type = mem_type;

	RegionAdd(&region);
	return TRUE;
}

/*
 * Find the first region that maps the given address in the current machine map
 */
MemRegion const*
MapAddr(ushort addr)
{
	MachineInitAll();

	int	r = RegionFind(addr);
	if (r >= 0)
		return machine->region+r;
	return 0;
}

/*
 * Initialisation of default machine maps.
 */

static	MemRegion	E9map[] =
{
	{ 0,      0x200,  &ram },		// RAM
	{ 0x1000, 0x103F, &ram },		// Registers
	{ 0x103F, 0x1040, &eeprom },		// Config register
	{ 0xB600, 0xB800, &eeprom },		// EEPROM
	{ 0xD000, 0x0000, &eprom }		// EPROM
};

static	MemRegion	E20map[] =
{
	{ 0,      0x300,  &ram },		// RAM
	{ 0x1000, 0x103F, &ram },		// Registers
	{ 0x103F, 0x1040, &eeprom },		// Config register
	{ 0x9000, 0xB000, &eprom2 },		// EPROM2
	{ 0xB600, 0xB800, &eeprom },		// EEPROM
	{ 0xD000, 0x0000, &eprom }		// EPROM
};

static	MemRegion	F1map[] =
{
	{ 0,      0x300,  &ram },		// RAM
	{ 0x1000, 0x103F, &ram },		// Registers
	{ 0x103F, 0x1040, &eeprom },		// Config register
	{ 0x1040, 0x1060, &ram },		// More Registers
	{ 0x300, 0x8000, &ram },		// external RAM
						// REVISIT: No flash map yet...
//	{ 0x8000, 0x0, &ram },			// external flash
	{ 0xFE00, 0xFF80, &eeprom }		// EEPROM: hide last 0x80
};

static	MemRegion	ExpMap[] =
{
	{ 0, 0x200,  &ram },			// RAM
	{ 0x1000, 0x103F,  &ram },		// Registers
	{ 0x103F, 0x1040, &eeprom },		// Config register
	{ 0x8000, 0x0, &ram }
};

static void
MachineInit(const char* name, MemRegion* regions, int num_region)
{
	int	i;

	MachineAdd(name);		// Add the machine name

	/*
	 * Add all regions, creating the memory type first if necessary
	 */
	for (i = 0; i < num_region; i++)
	{
		if (!MemTypeFind(regions[i].mem_type->name))
			MemTypeAdd(regions[i].mem_type);
		RegionAdd(regions+i);
	}
}

static void
MachineInitAll()
{
	if (num_machine != 0)
		return;
	MachineInit("E20", E20map, sizeof(E20map)/sizeof(MemRegion));
	MachineInit("E9", E9map, sizeof(E9map)/sizeof(MemRegion));
	MachineInit("F1", F1map, sizeof(F1map)/sizeof(MemRegion));
	MachineInit("Exp", ExpMap, sizeof(ExpMap)/sizeof(MemRegion));
}
