/*
 * $Id: midi.cxx,v 1.2 2000/07/29 07:45:42 cjh Exp $
 *
 * Display received MIDI data in readable ASCII form
 *
 * Copyright (c) 1996-2000, Clifford Heath.
 *
 * This program may be used for any commercial or noncommercial purpose
 * without charge, but acknowledgement of authorship and this copyright
 * message must be retained in any copies.
 */
#include	<stdincs.h>

#include	<util.h>
#include	<midi.h>
#include	<comms.h>

/*
 * The MidiStates indicate what's expected next.
 */
typedef	enum
{
	MidiReady,		/* Expecting anything */
	MidiNoteOn,		/* Expecting Note number to turn on */
	MidiNoteOnVelocity,	/* Expecting Note on velocity */
	MidiNoteOff,		/* Expecting Note number to turn off */
	MidiNoteOffVelocity,	/* Expecting Note off velocity (def 64) */
	MidiAfterTouch,		/* Expecting Note number for AfterTouch */
	MidiAfterTouchPressure,	/* Expecting AfterTouch Pressure */
	MidiControlChange,	/* Expecting Controller number */
	MidiControlValue,	/* Expecting Controller value */
	MidiProgramChange,	/* Expecting Program number */
	MidiChannelPressure,	/* Expecting Pressure value */
	MidiPitchWheel1,	/* Expecting 1st PitchWheel byte */
	MidiPitchWheel2,	/* Expecting 2nd PitchWheel byte */
	MidiSysEx,		/* System Exclusive Manufacturer Id */
	MidiSysExData		/* System Exclusive Data */
} MidiStateT;

char*	MidiControllers[] =
{
	"Bank Select",
	"Modulation Wheel (coarse)",
	"Breath controller (coarse)",
	"Foot Pedal (coarse)",
	"Portamento Time (coarse)",
	"Data Entry (coarse)",
	"Volume (coarse)",
	"Balance (coarse)",
	"Pan position (coarse)",
	"Expression (coarse)",
	"Effect Control 1 (coarse)",
	"Effect Control 2 (coarse)",
	"General Purpose Slider 1",
	"General Purpose Slider 2",
	"General Purpose Slider 3",
	"General Purpose Slider 4",
	"Bank Select (fine)",
	"Modulation Wheel (fine)",
	"Breath controller (fine)",
	"Foot Pedal (fine)",
	"Portamento Time (fine)",
	"Data Entry (fine)",
	"Volume (fine)",
	"Balance (fine)",
	"Pan position (fine)",
	"Expression (fine)",
	"Effect Control 1 (fine)",
	"Effect Control 2 (fine)",
	"Hold Pedal (on/off)",
	"Portamento (on/off)",
	"Sustenuto Pedal (on/off)",
	"Soft Pedal (on/off)",
	"Legato Pedal (on/off)",
	"Hold 2 Pedal (on/off)",
	"Sound Variation",
	"Sound Timbre",
	"Sound Release Time",
	"Sound Attack Time",
	"Sound Brightness",
	"Sound Control 6",
	"Sound Control 7",
	"Sound Control 8",
	"Sound Control 9",
	"Sound Control 10",
	"General Purpose Button 1 (on/off)",
	"General Purpose Button 2 (on/off)",
	"General Purpose Button 3 (on/off)",
	"General Purpose Button 4 (on/off)",
	"Effects Level",
	"Tremulo Level",
	"Chorus Level",
	"Celeste Level",
	"Phaser Level",
	"Data Button increment",
	"Data Button decrement",
	"Non-registered Parameter (coarse)",
	"Non-registered Parameter (fine)",
	"Registered Parameter (coarse)",
	"Registered Parameter (fine)",
	"All Sound Off",
	"All Controllers Off",
	"Local Keyboard (on/off)",
	"All Notes Off",
	"Omni Mode Off",
	"Omni Mode On",
	"Mono Operation",
	"Poly Operation"
};

struct {
	char*	name;
	char	id;
} MidiManufacturers[] =
{
	{ "Sequential Circuits",0x01 },
	{ "Big Briar",		0x02 },
	{ "Octave / Plateau",	0x03 },
	{ "Moog",		0x04 },
	{ "Passport Designs",	0x05 },
	{ "Lexicon",		0x06 },
	{ "Kurzweil",		0x07 },
	{ "Fender",		0x08 },
	{ "Gulbransen",		0x09 },
	{ "Delta Labs",		0x0A },
	{ "Sound Comp.",	0x0B },
	{ "General Electro",	0x0C },
	{ "Techmar",		0x0D },
	{ "Matthews Research",	0x0E },
	{ "Oberheim",		0x10 },
	{ "PAIA",		0x11 },
	{ "Simmons",		0x12 },
	{ "Gentle Electric",	0x13 },
	{ "Fairlight",		0x14 },
	{ "Peavey",		0x1B },
	{ "JL Cooper",		0x15 },
	{ "Lowery",		0x16 },
	{ "Lin",		0x17 },
	{ "Emu",		0x18 },
	{ "Bon Tempi",		0x20 },
	{ "S.I.E.L.",		0x21 },
	{ "SyntheAxe",		0x23 },
	{ "Hohner",		0x24 },
	{ "Crumar",		0x25 },
	{ "Solton",		0x26 },
	{ "Jellinghaus Ms",	0x27 },
	{ "CTS",		0x28 },
	{ "PPG",		0x29 },
	{ "Elka",		0x2F },
	{ "Kawai",		0x40 },
	{ "Roland",		0x41 },
	{ "Korg",		0x42 },
	{ "Yamaha",		0x43 },
	{ "Casio",		0x44 },
	{ "Akai",		0x45 }
};

void
DumpHexAndMidi(char ch)
{
	DumpHex(ch);
	DumpMidi(ch);
}

/*
 * Midi dumper. Gets a byte at a time, and uses static state to dump it.
 */
void
DumpMidi(char ch)
{
	static	MidiStateT	running_state = MidiReady;
	static	MidiStateT	state = MidiReady;
	static	int		channel = 0;	/* Channel number */
	static	int		first = 0;	/* First following byte */
	int			isstatus = (ch&0x80) != 0;
	int			i;

	/* Debug code:
	printf("[0x%02X,%d]", ch&0xFF, state);
	*/
	if (state < MidiSysEx && isstatus != (state == MidiReady))
	{
		if (isstatus)
		{
			/* New message starts, abandon old one */
			printf(
				"Unexpected end of midi message in state %d\n",
				state
			);
			state = MidiReady;
		}
		else if (running_state != MidiReady)
		{
			/* Non-status byte, assume running status */
			state = running_state;
		}
		else
		{
			/* Oops, can't do running status! */
			printf(
				"Running state %s with no previous state!\n",
				Ascii(ch)
			);
			return;
		}

	}

	switch (state)
	{
	case MidiReady:
		channel = ch&0xF;
		switch (ch&0xF0)
		{
		case 0x80:	state = MidiNoteOn; break;
		case 0x90:	state = MidiNoteOff; break;
		case 0xA0:	state = MidiAfterTouch; break;
		case 0xB0:	state = MidiControlChange; break;
		case 0xC0:	state = MidiProgramChange; break;
		case 0xD0:	state = MidiChannelPressure; break;
		case 0xE0:	state = MidiPitchWheel1; break;
		case 0xF0:		/* System */
			switch ((unsigned char)ch)
			{
			/*
			 * System common messages, reset running_state
			 */
			case 0xF0:	/* System Exclusive */
				running_state = MidiReady;
				printf("SysEx ");
				fflush(stdout);
				state = MidiSysEx;
				break;

			case 0xF1:	/* MTC Quarter Frame Message */
			case 0xF2:	/* Song Position Pointer */
			case 0xF3:	/* Song Select */
			case 0xF4:	/* Unassigned */
			case 0xF5:	/* Unassigned */
			case 0xF6:	/* Tune Request */
				/* Clear running state */
				running_state = MidiReady;
				return;

			case 0xF7:	/* End System Exclusive */
				printf("Stray End System Exclusive\n");
				running_state = MidiReady;
				state = MidiReady;
				return;

			/*
			 * Realtime messages
			 * Don't touch running_state!
			 */
			case 0xF8:	printf("MIDI Clock\n"); return;
			case 0xF9:	return;
			case 0xFA:	printf("MIDI Start\n"); return;
			case 0xFB:	printf("MIDI Continue\n"); return;
			case 0xFC:	printf("MIDI Stop\n"); return;
			case 0xFD:	return;
			case 0xFE:	printf("Active Sense\n"); return;
			case 0xFF:	printf("Reset\n"); return;
			}
		}
		running_state = state;
		return;

	case MidiNoteOn:
	case MidiNoteOff:
	case MidiAfterTouch:
	case MidiControlChange:
	case MidiPitchWheel1:
		state = (MidiStateT)((int)state+1); /* Next sequential state */
		first = ch;
		return;

	case MidiProgramChange:
		printf("ProgramChange %d\n", ch);
		break;

	case MidiChannelPressure:
		printf("ChannelPressure %d=%d\n", channel, ch);
		break;

	case MidiNoteOnVelocity:
		printf("NoteOn %d/%d=%d\n", channel, first, ch);
		break;

	case MidiNoteOffVelocity:
		printf("NoteOff %d/%d=%d\n", channel, first, ch);
		break;

	case MidiAfterTouchPressure:
		printf("AfterTouch %d/%d=%d\n", channel, first, ch);
		break;

	case MidiControlValue:
		printf(
			"Controller %d %s = %d\n",
			channel, MidiControllers[first], ch
		);
		break;

	case MidiPitchWheel2:
		printf("PitchWheel %d=%d\n", channel, first|(ch<<7));
		break;

	case MidiSysEx:
		if (isstatus)
		{
			printf("SysEx ended before it started!\n");
			break;
		}
		state = MidiSysExData;
		for (
			i = 0;
			i < (int)(sizeof(MidiManufacturers)/sizeof(char*));
			i++
		)
		{
			if (MidiManufacturers[i].id == ch)
			{
				printf(
					"(%s), \"",
					MidiManufacturers[(int)ch].name
				);
				fflush(stdout);
				return;
			}
		}
		printf("(mfg %d), \"", ch);
		return;

	case MidiSysExData:
		if (isstatus)
		{	/* Should be EndSysEx */
			printf("\"\n");	/* End System Exclusive */
			if ((unsigned char)ch != 0xF7)
			{
				printf(
					"System Exclusive ended badly (%s)\n",
					Ascii(ch)
				);
			}
			break;
		}
		printf("%s", Ascii(ch));	/* here's a data byte */
		return;				/* Continue SysEx state */
	}
	fflush(stdout);
	state = MidiReady;
}
