/*
 * $Id: serial.cxx,v 1.18 2002/06/10 03:26:12 cjh Exp $
 *
 * Serial device (tty) handler
 *
 * Copyright (c) 1996-2000, Clifford Heath.
 *
 * This program may be used for any commercial or noncommercial purpose
 * without charge, but acknowledgement of authorship and this copyright
 * message must be retained in any copies.
 */
#include	<stdincs.h>
#include	<util.h>
#include	<multiplex.h>
#include	<serial.h>

typedef	void	(*sighandler_t)(int);

#if defined(__unix)
static	speed_t	SpeedCode(int baud);
#endif

Serial::~Serial()
{
#if defined(__unix)
	if (fd < 0)
		return;
	tcflush(fd, TCIOFLUSH);
	tcsetattr(fd, TCSADRAIN, &orig);	// Restore modes.
#if defined(__linux)
	ioctl(fd, TIOCSSERIAL, &ss_orig);	// Reset special serial modes
#endif
	if (fd > 2)
		close(fd);
#elif defined(WIN32)
	CancelIo(comms);			// Cancel all async IO
	(void)SetCommState(comms, &orig);	// Restore modes
	CloseHandle(comms);			// Close comm port
#endif
}

Serial::Serial(const char* tty)
#if defined(__unix)
: fd(0)
#elif defined(WIN32)
: comms(0)
#endif
, max_delay(-1)
{
#if defined(__unix)
	if (tty)
		fd = open(tty, O_RDWR);
	if (fd < 0
	 || tcgetattr(fd, &orig) < 0)
	{
		perror(tty);
		if (fd >= 0 && tty)
			close(fd);
		fd = -1;
		return;
	}
	mode = orig;
#if defined(__linux)
	ioctl(fd, TIOCGSERIAL, &ss_orig);
	ss_mode = ss_orig;

	// start writing single bytes immediately, don't timeout waiting 4 more:
	ss_mode.flags |= ASYNC_LOW_LATENCY;

	if (ioctl(fd, TIOCSSERIAL, &ss_mode))
	{
		// This will print an error on a pty (like stdin under X)
		// perror("ASYNC_LOW_LATENCY");
	}
#endif
#elif defined(WIN32)
	/*
	 * set up for asynchronous I/O
	 */
	memset(&read_io, 0, sizeof(read_io));
	memset(&write_io, 0, sizeof(write_io));
	read_io.hEvent = CreateEvent(
			(LPSECURITY_ATTRIBUTES)0, // not Inheritable
			FALSE,			// Auto reset
			FALSE,			// not initially asserted
			"ComRead"
		);
	write_io.hEvent = CreateEvent(
			(LPSECURITY_ATTRIBUTES)0, // not Inheritable
			FALSE,			// Auto reset
			FALSE,			// not initially asserted
			"ComWrite"
		);

	/*
	 * Open the comm port
	 */
	comms = CreateFile(
			tty,				// file name
			GENERIC_READ|GENERIC_WRITE,	// access mode
			0,				// share mode
			(LPSECURITY_ATTRIBUTES)0,	// SD
			OPEN_EXISTING,			// how to create
			FILE_FLAG_OVERLAPPED,		// allows async io
			0				// handle to template
		);
	if (comms == 0)
		return;					// Error.

	/*
	 * Set the timeouts
	 */
	COMMTIMEOUTS	timeouts;			// Times in milliseconds
	timeouts.ReadIntervalTimeout = 0;		// max inter-char gap
	timeouts.ReadTotalTimeoutMultiplier = 0;	// max time per char
	timeouts.ReadTotalTimeoutConstant = 0;		// base to add to above
	timeouts.WriteTotalTimeoutMultiplier = 0;	// same for write (0)
	timeouts.WriteTotalTimeoutConstant = 0;		// same for write (0)
	/*BOOL*/SetCommTimeouts(
			comms,				// handle to comm device
			&timeouts			// time-out values
		);

	/*
	 * Clear outstanding I/O and set preferred queue lengths
	 */
	/*BOOL*/SetupComm(
			comms,		// handle to communications device
			512,		// size of input buffer
			512		// size of output buffer
		);

	/*
	 * Get the comport state
	 */
	/*BOOL*/GetCommState(
			comms,  // handle to communications device
			&orig    // device-control block
		);
	mode = orig;

	/* Query device driver for comm port capabilities
		COMMPROP comm_prop;		// communications properties
		BOOL GetCommProperties(
			comms,			// handle to comm device
			&comm_prop		// communications properties
		);
		if (comm_prop.dwMaxBaud < BAUD_9600)
			...
	*/

	/*
	 * Set the comport state
	 */
	mode.DCBlength = sizeof(mode);		// sizeof(DCB) 
	mode.fBinary = TRUE;			// binary mode, no EOF check 
	mode.fParity = FALSE;			// enable parity checking 
	mode.fOutxCtsFlow = FALSE;		// CTS output flow control 
	mode.fOutxDsrFlow = FALSE;		// DSR output flow control 
	mode.fDtrControl = DTR_CONTROL_ENABLE;	// DTR flow control type 
	mode.fDsrSensitivity = FALSE;		// DSR sensitivity 
	mode.fTXContinueOnXoff = FALSE;		// XOFF continues Tx 
	mode.fOutX = FALSE;			// XON/XOFF out flow control 
	mode.fInX = FALSE;			// XON/XOFF in flow control 
	mode.fErrorChar = FALSE;		// enable error replacement 
	mode.fNull = FALSE;			// enable null stripping 
	mode.fRtsControl = RTS_CONTROL_ENABLE;	// RTS flow control 
	mode.fAbortOnError = FALSE;		// abort on error 
	mode.XonLim = 0;			// transmit XON threshold 
	mode.XoffLim = 0;			// transmit XOFF threshold 
	mode.ByteSize = 8;			// number of bits/byte, 4-8 
	mode.Parity = 0;			// 0-4=no,odd,even,mark,space 
	mode.StopBits = 2;			// 0,1,2 = 1, 1.5, 2 
	mode.XonChar = 0;			// Tx and Rx XON character 
	mode.XoffChar = 0;			// Tx and Rx XOFF character 
	mode.ErrorChar = 0;			// error replacement character 
	mode.EofChar = 0;			// end of input character 
	mode.EvtChar = 0;			// received event character 

	/*
	mode.BaudRate;            // current baud rate 
	CBR_110 CBR_300 CBR_600 CBR_1200 CBR_2400 CBR_4800 CBR_9600
	CBR_14400 CBR_19200 CBR_38400 CBR_56000 CBR_57600 CBR_115200
	*/
	/*BOOL*/SetCommState(
			comms,
			&mode
		);

#if 0
	HasOverlappedIoCompleted(&overlapped);
	BOOL rf = ReadFile(
		comms,				// handle to file
		(void*)buffer,			// Don't touch buffer until done
		(DWORD)bytes_wanted,		// Number of bytes to read
		(DWORD*)bytes_read,		// Or (DWORD*)0, and check
		&read_io			// the overlapped I/O struct
	);
	BOOL WriteFile(
		comms,				// handle to file
		(void*)buffer,			// Don't touch buffer until done
		(DWORD)bytes_to_write,		// number of bytes to write
		(DWORD*)bytes_written,		// Or (DWORD*)0, and check
		&write_io			// the overlapped I/O struct
	);

	if (!rf && GetLastError() == ERROR_IO_PENDING)
		async io started

	BOOL PurgeComm(
			comms,  // handle to communications resource
		one of:
			PURGE_RXCLEAR		// Discard received input
			PURGE_RXABORT		// Same but abort async I/O
			PURGE_TXCLEAR		// Discard queued output
			PURGE_TXABORT		// Same but abort async I/O
	    );
#endif

#endif
}

boolean
Serial::isOk()
{
#if defined(__unix)
	return fd >= 0;
#elif defined(WIN32)
	return comms != 0;	// REVISIT for WIN32
#endif
}

/* Input modes: */

void
Serial::disableLineEdit(boolean disabled)
{
	/*
	 * Disabling LineEdit mode means we read one character at
	 * a time with no echo, but can still be interrupted etc.
	 */
#if defined(__unix)
	if (disabled)
		mode.c_lflag &= ~(ICANON|ECHO);
	else
		mode.c_lflag = orig.c_lflag;
	tcsetattr(fd, TCSADRAIN, &mode);
#elif defined(WIN32)
	// REVISIT for WIN32
#endif
}

void
Serial::disableSpecialInput(boolean disabled)
{
#if defined(__unix)
	/*
	 * Disabling Special Input modes means that we have a fully 8-bit
	 * clean data path with no interpretation or echo.
	 */
	if (disabled)
	{
		/* input modes */
		mode.c_iflag &= ~(IGNBRK|BRKINT|PARMRK|INPCK|ISTRIP|INLCR|IGNCR
#if	defined(IUCLC)
				|IUCLC
#endif
				|ICRNL|IXON|IXANY|IXOFF|IMAXBEL);
		mode.c_iflag |= IGNPAR;

		/* control modes; Odd parity ... */
		mode.c_cflag &= ~(CSIZE|PARENB|CRTSCTS|PARODD|HUPCL);
		mode.c_cflag |= CREAD|CS8|CSTOPB|CLOCAL;

		/* local modes */
		mode.c_lflag &= ~(ISIG|ICANON|IEXTEN|ECHO|FLUSHO
#if	defined(PENDIN)
				|PENDIN
#endif
			);
		mode.c_lflag |= NOFLSH;
	}
	else
	{
		mode.c_iflag = orig.c_iflag;
		mode.c_cflag = orig.c_cflag;
		mode.c_lflag = orig.c_lflag;
	}

	tcsetattr(fd, TCSADRAIN, &mode);
#elif defined(WIN32)
	// REVISIT for WIN32
#endif
}

size_t
Serial::read(unsigned char* buf, size_t max, size_t min)
{
	size_t	total = 0;
	int	r;

	if (min == 0)
	{		// Non-blocking I/O
		Multiplexed*	mpx_pp = this;
		long		prev_delay = max_delay;
		max_delay = 10;

		Multiplex(&mpx_pp, 1);
		max_delay = prev_delay;
		if (!isReady())
			return 0;
	}

	is_ready = FALSE;
#if defined(__unix)
	while ((r = ::read(fd, buf, max)) >= 0
	 || errno == EINTR)	// Interrupted (SIGTSTP?), continue
	{
#if defined(TRACE_MCU_SERIAL)
	if (fd > 2)
	{
		printf("Read from MCU:");
		for (int i = 0; i < r; i++)
			printf(" 0x%02X", buf[i]);
		printf("\n");
	}
#endif
		total += r;
		if (total >= min)
			return total;
		buf += r;
		max -= r;
		min -= r;
	}
	if (total == 0 && r < 0)
		return r;	// Return an error only if we got no data first
#elif defined(WIN32)
	// REVISIT for WIN32
#endif

	return total;
}

void
Serial::discard()
{
#if defined(__unix)
	tcflush(fd, TCIFLUSH);
#elif defined(WIN32)
	// REVISIT for WIN32
#endif
}

void
Serial::write(const unsigned char* buf, size_t count)
{
#if defined(TRACE_MCU_SERIAL)
	if (fd > 2)
	{
		printf("Writing to MCU:");
		for (size_t i = 0; i < count; i++)
			printf(" 0x%02X", buf[i]);
		printf("\n");
	}
#endif
#if defined(__unix)
	::write(fd, buf, count);	// Ignore write errors
#elif defined(WIN32)
	DWORD	written;
	if (WriteFile(comms, buf, count, &written, &write_io))
		return;			// Complete and ok

	// I/O incomplete? wait for it.
	if (GetLastError() == ERROR_IO_PENDING)
		(void)WaitForSingleObject(write_io.hEvent, 30); // 30ms max

	//(void)WaitForMultipleObject(count, hEvent[], FALSE/*any*/, 30);
#endif
}

void
Serial::drain()
{
#if defined(__unix)
	tcdrain(fd);
#elif defined(WIN32)
	// REVISIT for WIN32
#endif
}

/* Output modes: */
void
Serial::disableSpecialOutput(boolean disabled)
{
#if defined(__unix)
	if (disabled)
		mode.c_oflag &= ~OPOST;
	else
		mode.c_oflag = orig.c_oflag;

	tcsetattr(fd, TCSADRAIN, &mode);
#elif defined(WIN32)
	// REVISIT for WIN32
#endif
}

boolean
Serial::setSpeed(int baud)
{
#if defined(__unix)
	speed_t	baudcode = SpeedCode(baud);

	if (!baudcode)
		return FALSE;		// This speed is impossible!

#if defined(TRACE_MCU_SERIAL)
	if (fd > 2)
		printf("Setting MCU baud rate to %d\n", baud);
#endif

#if defined(__linux) && defined(TIOCSSERIAL)
	// odd baud-rate support:
	if (baudcode == (speed_t)-1)
	{
		ss_mode.flags |= ASYNC_SPD_CUST;
		ss_mode.custom_divisor = ss_mode.baud_base/baud;

		baudcode = B38400;
	}
	else
		ss_mode.flags &= ~ASYNC_SPD_CUST;

	if (ioctl(fd, TIOCSSERIAL, &ss_mode))
	{
		perror("ASYNC_SPD_CUST");
	}
#endif
	// Normal baud-rate setting:
	baud_rate = baud;

	cfsetispeed(&mode, baudcode);
	cfsetospeed(&mode, baudcode);
	tcsetattr(fd, TCSADRAIN, &mode);
#elif defined(WIN32)
	// REVISIT for WIN32
#endif
	return TRUE;
}

int
Serial::getSpeed()
{
	return baud_rate;
}

/*
 * If catching SIGTSTP, callers should:
 * - call suspend();
 * - execute kill(getpid(), SIGTSTP);		// Stop ourselves
 * - Reinstate the signal handler: signal(SIGTSTP, (sighandler_t)suspend);
 * - call resume();
 */
void
Serial::suspend()	// Call from SIGTSTP interrupt handler
{
#if defined(__unix)
#if	defined(SIGTSTP)
	if (fd == 0)
	{		// Stdin needs special processing
		signal(SIGTTOU, (sighandler_t)SIG_IGN);	// Prevent output prob.
		tcsetattr(fd, TCSADRAIN, &orig);	// Restore cooked mode
#if defined(__linux)
		ioctl(fd, TIOCSSERIAL, &ss_orig);	// Reset special serial
#endif
		putchar('\n');				// Start of new line
		signal(SIGTTOU, (sighandler_t)SIG_DFL);	// Back to normal
		return;
	}
#endif	/* SIGTSTP */

	tcsetattr(fd, TCSADRAIN, &mode);	// Restore cooked mode
#if defined(__linux)
	ioctl(fd, TIOCSSERIAL, &ss_mode);	// Reset special serial
#endif
#elif defined(WIN32)
	// REVISIT for WIN32
#endif
}

void
Serial::resume()
{
#if defined(__unix)
	tcsetattr(fd, TCSADRAIN, &mode);	/* Restore current mode */
#elif defined(WIN32)
	// REVISIT for WIN32
#endif
}

/* functions inherited from Multiplexed: */
boolean
Serial::request(	// Is a request required?
	int*		a_fd,		// fd to poll
	boolean*	is_output,	// read or write
	long*		a_max_delay	// in millisecs
)
{
#if defined(__unix)
	is_ready = FALSE;

	*a_fd = fd;			// Poll this fd
	*is_output = FALSE;		// for input
	*a_max_delay = max_delay;	// delaying not more than max_delay ms
	return TRUE;			// We want a poll request done
#elif defined(WIN32)
	is_ready = HasOverlappedIoCompleted(&read_io);
	return FALSE;			// No fd to do a poll on
#endif
}

void
Serial::ready()		// Multiplexer found it ready
{
	is_ready = TRUE;
}

long
Serial::setMaxDelay(long millis)
{
	long	old_max = max_delay;
	max_delay = millis;
	return old_max;
}

#if defined(__unix)
static	speed_t
SpeedCode(int baud)
{
	switch (baud)
	{
	case 300:	return B300;
	case 600:	return B600;
	case 1200:	return B1200;
	case 2400:	return B2400;
	case 4800:	return B4800;
	case 9600:	return B9600;
	case 19200:	return B19200;
	case 38400:	return B38400;
	case 57600:	return B57600;
#if	defined(B115200)
	case 115200:	return B115200;
#endif
#if	defined(B230400)
	case 230400:	return B230400;
#endif

#if defined(__linux) && defined(TIOCSSERIAL)
			// Allow any speed using custom divisor
	default:	return (speed_t)-1;
#else
	default:	return 0;
#endif
	}
}
#endif
