/*
 * $Id: symbol.cxx,v 1.3 2001/02/06 11:14:24 cjh Exp $
 *
 * Symbol table routines
 *
 * A very simple sorted array of pointers to Symbol structures...
 *
 * Copyright (c) 1996-2000, Clifford Heath.
 *
 * This program may be used for any commercial or noncommercial purpose
 * without charge, but acknowledgement of authorship and this copyright
 * message must be retained in any copies.
 */
#include	<stdincs.h>

#include	<util.h>
#include	<db11.h>
#include	<symbol.h>

class Symbol
{
public:
	Symbol(const char* name, ulong value, SymState state);
	~Symbol();

	char*		name;
	ulong		value;
	SymState	state;
};

static	Symbol**	symtab;
static	int		sym_count;
static	int		sym_alloc;

static	void		SymbolSort();
static	int		SymbolFind(const char* name);
static	int		SymbolFind(ulong value, SymState);
static	void		SymbolGrow();

void
SymbolDefine(		// (un)define named symbol to value
	const char*	name,
	ulong		value,
	SymState	state
)
{
	Symbol*		sym;
	int		i = SymbolFind(name);

	if (state == SymUndef)
	{		// Undefine the symbol
		if (i < 0)
			return;		// Wasn't defined anyhow

		// Shuffle the table down.
		sym = symtab[i];
		delete sym;
		memmove(symtab+i, symtab+i+1, (sym_count-i-1)*sizeof(Symbol*));
		return;
	}

	// REVISIT: This doesn't allow definition of case-clashing symbols!

	if (i < 0)
	{		// Not found, make room
		SymbolGrow();
		i = sym_count++;
	}
	else
		delete symtab[i];	// Delete old symbol
	symtab[i] = new Symbol(name, value, state);

	SymbolSort();
}

SymState
SymbolValue(		// Get value of named symbol
	const char*	name,
	ulong&		value
)
{
	Symbol*		sym;
	int		i = SymbolFind(name);

	if (i < 0)
		return SymUndef;

	sym = symtab[i];
	value = sym->value;	// Return the value
	return sym->state;	// And the state
}

SymState
SymbolNear(		// Find code symbol closest before addr
	ulong		addr,
	char*&		name,
	SymState	state
)
{
	Symbol*		sym;
	int		i = SymbolFind(addr, state);

	if (i < 0)
		return SymUndef;

	sym = symtab[i];
	name = strdup(sym->name);	// Return the value
	return sym->state;		// And the state
}

Symbol::Symbol(const char* _name, ulong _value, SymState _state)
: name(strdup(_name))
, value(_value)
, state(_state)
{
	// Nothing to do.
}

Symbol::~Symbol()
{
	delete name;
}

static int
SymbolCompare(const void* a, const void* b)
{
	return (*(Symbol**)a)->value - (*(Symbol**)b)->value;
}

static void
SymbolSort()
{
	qsort(symtab, sym_count, sizeof(Symbol*), SymbolCompare);
}

static int
SymbolFind(const char* name)
{
	int	closest = -1;

	for (int i = 0; i < sym_count; i++)
	{
		if (strcmp(symtab[i]->name, name) == 0)
			return i;
		if (strcasecmp(symtab[i]->name, name) == 0)
			closest = i;
	}
	return closest;	// No exact match, return last case-independent match
}

static int
SymbolFind(ulong value, SymState state)
{
        register int    hi, lo, mid;

        lo = 0;
        hi = sym_count-1;
        while (hi >= lo)
        {
                mid = (lo+hi)/2;
                if (value < symtab[mid]->value)
                        hi = mid-1;
                else if (value > symtab[mid]->value)
                        lo = mid+1;
                else
		{
			lo = mid;
			break;
		}
        }
	if (lo >= 0 && (lo == sym_count || symtab[lo]->value > value))
		lo--;

	// If we wanted a particular kind of symbol, search backwards for it:
	while (lo >= 0
	    && state != SymUndef
	    && symtab[lo]->state != state)
		lo--;
        return lo;
}

/*
 * Make enough room to add one symbol
 */
static void
SymbolGrow()
{
	if (sym_count+1 < sym_alloc)
		return;

	// Allocate more memory and copy the array.
	sym_alloc = sym_alloc+64;
	Symbol**	tab;
	tab = new Symbol*[sym_alloc];
	memcpy(tab, symtab, sym_count*sizeof(Symbol*));
	delete symtab;
	symtab = tab;
}

#if	defined(TEST)
main()
{
	char		buf[256];
	ulong		addr;
	ulong		addr2;
	SymState	state;
	char*		name;

	while (printf("> "), gets(buf))
	{
		if (!isatty(0))
			puts(buf);

		if (buf[0] == '?')	// Search
		{
			if (isdigit(buf[1]))
			{		// Numeric search
				addr = strtol(buf+1, (char**)0, 0);
				state = SymbolNear(addr, name);
				if (state != SymUndef)
				{
					SymbolValue(name, addr2);
					printf("Found %s symbol %s at %ld\n",
						state == SymCode
							? "code"
							: "data",
						name,
						addr2
					);
					delete name;
				}
				else
					printf("No symbol found\n");
			}
			else
			{
				name = buf+1;
				state = SymbolValue(name, addr2);
				if (state != SymUndef)
				{
					printf("%s=%ld (%s)\n",
						name,
						addr2,
						state == SymCode
							? "code"
							: "data"
					);
				}
				else
					printf("symbol %s not found\n", name);
			}
		}
		else if ((name = strchr(buf, '=')) != 0)
		{			// Define
			*name++ = '\0';
			addr = strtol(name, (char**)0, 0);
			SymbolDefine(buf, addr);
		}
		else if ((name = strchr(buf, '@')) != 0)
		{			// Define
			*name++ = '\0';
			addr = strtol(name, (char**)0, 0);
			SymbolDefine(buf, addr, SymData);
		}
		else
			printf("Huh?\n");
	}
	putchar('\n');
}
#endif	/* TEST */
