/*
 * $Id: talker.cxx,v 1.23 2002/04/01 13:02:14 cjh Exp $
 *
 * Low-level Talker routines.
 *
 * Copyright (c) 1996-2000, Clifford Heath.
 *
 * This program may be used for any commercial or noncommercial purpose
 * without charge, but acknowledgement of authorship and this copyright
 * message must be retained in any copies.
 */
#include	<stdincs.h>

#include	<ctype.h>

#include	<util.h>
#include	<comms.h>
#include	<db11.h>
#include	<serial.h>
#include	<talker.h>
#include	<map.h>

#define	STOPPED	'\003'

/*
 * Stack frame, as transmitted:
 */
typedef	struct
{
	uchar	ccr;
	uchar	accb;
	uchar	acca;
	uchar	xHi;
	uchar	xLo;
	uchar	yHi;
	uchar	yLo;
	uchar	pcHi;
	uchar	pcLo;
} RawFrame;

/*
 * Copies of talker variables:
 */
ushort		eaddr = 0xFFFF;		// Address to get/put
boolean		program_all = TRUE;	// Set all programming registers
ushort		ereg;			// Address of programming register
uchar		eprep;			// Prog Reg value before latching data
uchar		estart;			// Prog Reg value to program
uchar		edone;			// Prog Reg value to finish programming

/*
 * Last stack frame value we received (where regs are saved)
 */
ushort		stack;
enum {					// Is the talker stopping?
	ST_Run, ST_Stopping, ST_StackLo, ST_Stopped
} stopping_state = ST_Stopped;

/*
 * Whether we last asked for the external programming voltage to be on or off:
 */
boolean		voltage_on = FALSE;	// External programming voltage

/*
 * Read a byte from the talker, timing out and returning EOF after 0.1 seconds.
 */
int
TalkerGet1Soon()
{
	Multiplexed*	mpx_p = mcu;
	uchar		ch;

	// TraceTime("Get1Soon\n");

	mcu->setMaxDelay(100);
	Multiplex(&mpx_p, 1);
	if (!mcu->isReady())
	{
		TraceTime("Got timeout\n");
		return EOF;
	}
	if (mcu->read(&ch, 1) == 1)
	{
		TraceTime("Got 0x%02X\n", ch);
		return ch;
	}
	TraceTime("Got read error\n");
	return EOF;
}

/*
 * The talker has just started up, synchronise with it.
 */
boolean
TalkerSync()
{
	uchar		buf[2];
	int		got;
	boolean		grace;

	sleep(1);
	program_all = TRUE;	// We don't know the actual values
	grace = TRUE;		// Allow a null byte response the first time
	for (int tries = 0; tries < 3; tries++)
	{
		/*
		 * If the talker is already synchronised, this will be
		 * interpreted as an illegal command and elicit an SWI
		 * reponse, in which case we should ignore the echo and
		 * expect the "STOPPED" response anyway.
		 */
		mcu->write((uchar*)"\0", 1);

	grace_retry:
		if ((got = TalkerGet1Soon()) == EOF)
			continue;
		if (got != STOPPED)
		{
			if (grace && got == 0)
			{		// Was already synchronised.
				grace = FALSE;
				goto grace_retry;
			}
			if (got != '\0')
			{
				printf("Talking garbage: ");
				DumpAscii(got);
				printf("\n");
			}
			continue;
		}
		if ((got = TalkerGet1Soon()) == EOF)
			goto timeout;
		buf[0] = got;
		if ((got = TalkerGet1Soon()) == EOF)
		{
	timeout:	printf("Timeout, trying again\n");
			continue;
		}
		buf[1] = got;
		eaddr = buf[0]+(buf[1]<<8);
		stack = eaddr;
		stopping_state = ST_Stopped;
		printf("Talking, stack pointer at 0x%0X\n", eaddr);
		return TRUE;
	}
	printf("Talker synchronisation failed\n");
	return FALSE;
}

/*
 * Send a byte to the Talker, and expect it to be echoed verbatim.
 */
boolean
TalkerSendCheckEcho(uchar byte)
{
	int	echo;

	TraceTime("SendCheckEcho 0x%02X\n", byte);
	mcu->write(&byte, 1);
	if ((echo = TalkerGet1Soon()) == EOF)
	{
		printf("No echo from 0x%02x\n", byte);
		return FALSE;
	}
	else if ((uchar)echo != byte)
	{
		printf("Invalid echo 0x%02x from 0x%02x\n", echo, byte);
		return FALSE;
	}
	return TRUE;
}

/*
 * Set the current Get/Put address (EADDR) in the talker:
 */
boolean
TalkerSetAddr(short addr)
{
	if (!TalkerSendCheckEcho('\x39')
	 || !TalkerSendCheckEcho(addr>>8)
	 || !TalkerSendCheckEcho('\x3A')
	 || !TalkerSendCheckEcho(addr))
	{
		printf("Failed to set address 0x%04X for operation\n", addr);
		return FALSE;
	}
	eaddr = addr;
	return TRUE;
}

/*
 * Read the specified number of bytes from the current address.
 * Return number of bytes read.
 */
int
TalkerGetMem(uchar* buf, int count)
{
	uchar		getcmd;
	int		byte;
	int		total = 0;
	int		r = 0;

	// Handle large buffers:
	while (count > 64)
	{
		if ((r = TalkerGetMem(buf, 64)) != 64)
			return total+r;
		buf += 64;
		count -= 64;
		total += 64;
	}

	// Send command:
	getcmd = 0x40+count-1;
	if (!TalkerSendCheckEcho(getcmd))
	{
		printf("Get memory command not properly acknowledged\n");
		return r;
	}

	// Get returned bytes:
	while (count-- > 0)
	{
		if ((byte = TalkerGet1Soon()) == EOF)
		{
			printf("Lost synchronisation reading memory\n");
			return total;
		}
		eaddr++;
		total++;
		*buf++ = byte;
	}
	return total;
}

boolean
TalkerVerifyEcho(ushort addr, uchar expect)
{
	int	echo;

	// Verify byte
	if ((echo = TalkerGet1Soon()) == EOF)
	{
		printf("At 0x%04X, no echo from 0x%02x\n", addr, expect);
		return FALSE;
	}
	else if ((uchar)echo != expect)
	{
		printf(
			"At 0x%04X, invalid echo 0x%02X from 0x%02X\n",
			addr, echo, expect
		);
		return FALSE;
	}
	return TRUE;
}

/*
 * Write the bytes passed into the current address,
 * returning how many were written.
 */
int
TalkerPutMem(uchar* buf, int count)
{
	uchar		putcmd;
	int		total = 0;
	int		r = 0;
	int		fast_verified = 0;

#if 0
	printf("Putting %d bytes to 0x%04X\n", count, eaddr);
	for (int i = 0; i < count; i++)
		printf("%02X ", buf[i]);
	printf("\n");
#endif

	// Handle large buffers
	while (count > 64)
	{
		if ((r = TalkerPutMem(buf, 64)) != 64)
			return total+r;
		buf += 64;
		count -= 64;
		total += 64;
	}

	if (not_really || count <= 0)
		return count;

	if (SLOW_MODE)
	{
		// Send command:
		putcmd = 0x80+count-1;
		TraceTime("SendCheck Put command (%d bytes)\n", count);
		if (!TalkerSendCheckEcho(putcmd))
		{
			printf("Put memory command not acknowledged\n");
			return r;
		}

		/*
		 * Transmit up to 64 bytes, and verify all.  The last
		 * byte is verified by a Get1 command as Put doesn't
		 * echo the last byte.
		 */
		TraceTime("put byte 0x%02X\n", *buf);
		mcu->write(buf++, 1);		// Transmit first byte
		TraceTime("byte gone ok\n");
		eaddr++;
		while (--count > 0)
		{
			TraceTime("put byte 0x%02X\n", *buf);
			mcu->write(buf++, 1);	// Transmit next byte
			TraceTime("byte gone ok\n");
			eaddr++;

			// Verify previous byte (not the one just transmitted)
			TraceTime("Verify echo:\n");
			if (!TalkerVerifyEcho(eaddr-2, buf[-2]))
				break;

			total++;		// Count ok bytes
		}
		uchar	getcmd = 0x40;	// Send GetMem to verify last byte
		if (!TalkerSendCheckEcho(getcmd))
			goto bad;
		if (count == 0)			// Verify final echo
		{
			TraceTime("Verify final echo:\n");
			if (!TalkerVerifyEcho(eaddr-1, buf[-1]))
				goto bad;
			total++;
		}
		else if (count > 0)
		{
		bad:	printf("Lost synchronisation writing memory\n");
			return total;
		}
	}
	else
	{
		/*
		 * Fast mode PutMem: doesn't check the acknowledgement
		 * of each byte as it's sent.
		 */
		unsigned char	sendbuf[66];	// Put+64+Get1
		unsigned char	recvbuf[66];	// PutAck+63+GetAck+1
		boolean		problems = FALSE;
		int		expected = count+2;	// bytes expected back

		/*
		 * Send command and data in one write.
		 */
		sendbuf[0] = 0x80+count-1;	// PutMem command
		memcpy(sendbuf+1, buf, count);	// Data
		sendbuf[expected-1] = 0x40;	// Get1 command
		mcu->write(sendbuf, expected);

		/*
		 * Read response.  Allow time for 24 characters to be echoed.
		 * A bit of extra time doesn't hurt because we expect to stop
		 * when we get the last bytes anyhow.
		 */
		mcu->setMaxDelay(24*10000L/mcu->getSpeed() + 10);

		/*
		 * We get woken up when the FIFO's nearly full,
		 * so we won't read the whole amount in one go!
		 */
		while (fast_verified < expected
		    && (r = mcu->read(recvbuf+fast_verified, expected-fast_verified)) > 0)
			fast_verified += r;

		/*
		 * Because PutMem echoes all but the last byte, we Get1 which
		 * is echoed. The echo is checked and removed here.
		 */
		if (recvbuf[fast_verified-2] == 0x40)
		{
			// shift remaining byte down for ease below
			recvbuf[fast_verified-2] = recvbuf[fast_verified-1];
			fast_verified--;
		}
		else
		{
			printf("Get1 for final verify on PutMem not acknowledged correctly\n");
			problems = TRUE;
			if (fast_verified > count+1)
				// Prevent comparison buffer overrun.
				fast_verified = count+1;
		}

		total += fast_verified;

		/*
		 * Check all the echoed data we received, then the length:
		 * The 2nd last byte we receive should be 0x40, acknowledging
		 * the Get1 which verifies the last byte.
		 */
		for (int i = 0; i < fast_verified; i++)
		{
			if (sendbuf[i] == recvbuf[i])
				continue;
			if (i == 0)
				printf("PutMem command acknowledged incorrectly\n");
			else
				printf("PutMem 0x%02X to address 0x%04X "
					"acknowledged incorrectly as 0x%02X\n",
					sendbuf[i],
					eaddr+i-1,
					recvbuf[i]
				);
			if (problems)
			{
				int j;
				printf("...further problems skipped\n");
				printf("fast_verified = %d\n", fast_verified);
				for (j = 0; j < fast_verified; j++)
					printf("%02x", sendbuf[j]);
				printf("\n");
				for (j = 0; j < fast_verified; j++)
					printf("%02x", recvbuf[j]);
				printf("\n");
				break;
			}
			problems = TRUE;
		}
		if (total == 0)
			printf("No response to PutMem command\n");
		else if (total < count+1)
			printf(
				"Short response to PutMem (got %d of %d)\n",
				total, count+1
			);
		if (total > 0)
			total--;	// Exclude command byte.
		eaddr += total;
	}
	return total;
}

/*
 * Program the bytes passed, into the current address,
 * returning how many were programmed.
 * The correct programming variable values must have been set up first,
 * and the memory erased if necessary.
 */
int
TalkerProgMem(uchar* buf, int count, int delay_ms)
{
	uchar		progcmd;
	int		echo;
	int		total = 0;
	int		r = 0;

	if (not_really)
		return count;

	// Handle large buffers
	while (count > 64)
	{
		if ((r = TalkerProgMem(buf, 64, delay_ms)) != 64)
			return total+r;
		buf += 64;
		count -= 64;
		total += 64;
	}

	// Send command:
	progcmd = 0xC0+count-1;
	if (!TalkerSendCheckEcho(progcmd))
	{
		printf("Program memory command not properly acknowledged\n");
		return r;
	}

	/*
	printf(
		"Using %d+1 delay chars for %d ms\n",
		delay_ms*mcu->getSpeed() / (8*1000),
		delay_ms
	);
	*/

	// Transmit and verify the bytes:
	while (count > 0)
	{
		// printf("Programming 0x%02X\n", *buf);
		mcu->write(buf, 1);

		// Output a series of Null characters then FF to form time delay
		for (
			int idle = 0;
			idle < delay_ms*mcu->getSpeed() / (8*1000);
			idle++
		)
			mcu->write((uchar*)"\0", 1);
		mcu->write((uchar*)"\xFF", 1);

		if ((echo = TalkerGet1Soon()) == EOF)
		{
			printf("Lost synchronisation programming memory, RESET NOW\n");
			return total;
		}
		else if ((uchar)echo != *buf)
		{
			printf("Programming memory at 0x%04X failed (0x%02x instead of 0x%02x)\n", eaddr, echo, *buf);
			total--;	// Keep going, but lower success count
		}
		/*
		else
			printf("Programmed Ok\n");
		*/
		eaddr++;
		buf++;
		count--;
		total++;
	}
	return total;
}

static void
Voltage(boolean on)
{
	uchar	buf[256];

	if (voltage_on != on)
	{
		printf("%s external programming voltage and press ENTER: ",
			voltage_on ? "Remove" : "Apply"
		);
		fflush(stdout);
		stdio->read(buf, sizeof(buf));

		voltage_on = on;
	}
}

/*
 * Write (E)EPROM in one region only, erasing first if needed/possible.
 */
static int
PromWrite(
	MemRegion const*	region,
	ushort			addr,
	uchar*			data,
	int			size,
	boolean			erase_first
)
{
	MemType const*		mem_type = region->mem_type;

	Voltage(mem_type->external_voltage);

	if (erase_first && mem_type->prog_reg)
	{
		/*
		 * REVISIT: We could read the memory block
		 * here and decide whether erasure is needed...
		 */

		/*
		 * This only uses byte-erase, no block erase:
		 */
		if (mem_type->prep_erase == 0
		 && mem_type->start_erase == 0)
		{
			printf("Memory at 0x%04X not erasable, skipping erase step (take your chances)\n", addr);
		}
		else
		{
			uchar	buf[256];
			int	bytes = size;

			memset(buf, 0xFF, sizeof(buf));
	TraceTime("Setting vars for erase\n");
			if (!TalkerSetProgVars(
					mem_type->prog_reg,
					mem_type->prep_erase,
					mem_type->start_erase,
					mem_type->end_erase
				)
			)
				return 0;

			if (eaddr != addr)
			{
				if (!TalkerSetAddr(addr))
					return 0;
			}
			while (bytes > (int)sizeof(buf))
			{
				if (TalkerProgMem(
						buf, sizeof(buf),
						mem_type->delay
					) != sizeof(buf)
				)
					goto erase_fail;
			}
			if (bytes > 0)
			{
				if (TalkerProgMem(
						buf, bytes,
						mem_type->delay
					) != bytes
				)
				{
				erase_fail:
					printf("Erase failed, giving up\n");
					return 0;
				}
			}
		}
	}
	// else printf("No erase needed\n");

	// printf("PromWrite: prog_reg = 0x%04X\n", mem_type->prog_reg);

	TraceTime("Setting vars for (e)eprom write\n");
	if (!TalkerSetProgVars(
			mem_type->prog_reg,
			mem_type->prep_prog,
			mem_type->start_prog,
			mem_type->end_prog
		)
	)
		return 0;
	if (eaddr != addr)
	{
		if (!TalkerSetAddr(addr))
			return 0;
	}
	return TalkerProgMem(data, size, mem_type->delay);
}

/*
 * Write or program a memory block, using the memory map to decide how.
 */
int
TalkerWrite(ushort addr, uchar* data, int count, boolean erase_first)
{
	MemRegion const*	region;
	int			total = 0;
	int			size;
	int			done;

	while (count > 0)
	{
		boolean		is_rom;

		/* Find which memory region we're in */
		if ((region = MapAddr(addr)) == 0)
		{
			printf(
				"Can't write at 0x%04X outside mapped memory\n",
				addr
			);
			break;
		}

		/* Work out how much we can do in one call */
		size = count;
		if (size >= ((region->end_addr - addr) & 0xFFFF))
			size = (region->end_addr - addr) & 0xFFFF;

		is_rom = region->mem_type && region->mem_type->prog_reg;
		if (not_really || verbose)
			printf("%s %d bytes of memory at 0x%04x\n",
				is_rom ? "Program" : "Write",
				size,
				addr
			);

		/* Write or program the memory */
		if (!is_rom)
		{
			if (eaddr != addr)
				TalkerSetAddr(addr);
			done = TalkerPutMem(data, size);
		}
		else
			done = PromWrite(region, addr, data, size, erase_first);

		total += done;
		addr += size;
		count -= size;
	}
	Voltage(FALSE);
	return total;
}

boolean
TalkerSetProgVars(
	ushort	prog_reg, // Address of the programming control register
	uchar	prepare,  // Value to store in PCR before programming
	uchar	start,	  // Value to store in PCR to start programming
	uchar	done	  // Value to store in PCR to end programming
)
{
	// printf("Program all o%s\n", program_all ? "n" : "ff");

	if ((prog_reg != ereg || program_all)
	 	&& (!TalkerSendCheckEcho('\x3B')
		 || !TalkerSendCheckEcho(prog_reg>>8)
		 || !TalkerSendCheckEcho('\x3C')
		 || !TalkerSendCheckEcho(prog_reg))
	 || (prepare != eprep || program_all)
	 	&& (!TalkerSendCheckEcho('\x3D')
		 || !TalkerSendCheckEcho(prepare))
	 || (start != estart || program_all)
	 	&& (!TalkerSendCheckEcho('\x3E')
		 || !TalkerSendCheckEcho(start))
	 || (done != edone || program_all)
		&& (!TalkerSendCheckEcho('\x3F')
		 || !TalkerSendCheckEcho(done)))
	{
		printf("Failed to set programming variables\n");
		program_all = TRUE;
		return FALSE;
	}
	program_all = FALSE;
	ereg = prog_reg;
	eprep = prepare;
	estart = start;
	edone = done;

	/*
	printf("Programming vars set to 0x%04X, 0x%02X, 0x%02X, 0x%02X\n",
		prog_reg, prepare, start, done
	);
	*/
	return TRUE;
}

boolean
TalkerGetRegisters(Frame* frame)
{
	uchar		trash;		// Stack pointer post-decrements
	RawFrame	raw;

	memset(&raw, 0xBA, sizeof(raw));
	if (eaddr == stack		// address is often already set.
	 && TalkerGetMem(&trash, 1) != 1)
	 	return FALSE;
	else if (!TalkerSetAddr(stack+1))
		return FALSE;

	if (TalkerGetMem((uchar*)&raw, 9) != 9)
	 	return FALSE;
	frame->sp = stack+9;		// Show stack *before* breakpoint
	frame->ccr = raw.ccr;
	frame->acca = raw.acca;
	frame->accb = raw.accb;
	frame->x = (raw.xHi << 8) + raw.xLo;
	frame->y = (raw.yHi << 8) + raw.yLo;
	frame->pc = (raw.pcHi << 8) + raw.pcLo;
	return TRUE;
}

boolean
TalkerSetRegisters(Frame* frame)
{
	RawFrame	raw;

	stack = frame->sp - 9;	// Frame contains stack *before* breakpoint

	memset(&raw, 0xBA, sizeof(raw));
	if (!TalkerSetAddr(stack+1))
	 	return FALSE;

	raw.ccr = frame->ccr;
	raw.acca = frame->acca;
	raw.accb = frame->accb;
	raw.xHi = frame->x >> 8;
	raw.xLo = frame->x;
	raw.yHi = frame->y >> 8;
	raw.yLo = frame->y;
	raw.pcHi = frame->pc >> 8;
	raw.pcLo = frame->pc;

	return TalkerPutMem((uchar*)&raw, 9) == 9;
}

/*
 * Setting the stack is unlike setting other registers.
 * We save it here instead of in the talker.
 */
boolean
TalkerSetStack(ushort val)
{
	stack = val-9;	// Outside world thinks about pre-breakpoint stack
	return TRUE;
}

boolean
processTalkerInput()
{
	unsigned char	buf[256];
	int		r;
	int		i;

	r = mcu->read(buf, sizeof(buf));
	for (i = 0; i < r; i++)
	{
		switch (stopping_state)
		{
		case ST_Run:
			if (buf[i] == STOPPED)
				stopping_state = ST_Stopping;
			else
				Dump((char)buf[i]);
			break;

		case ST_Stopping:
			stack = buf[i];
			stopping_state = ST_StackLo;
			break;

		case ST_StackLo:
			stack |= buf[i]<<8;
			stopping_state = ST_Stopped;
			break;

		case ST_Stopped:	// Shouldn't happen
			Dump((char)buf[i]);
			break;
		}
	}
	return stopping_state != ST_Stopped;
}

/*
 * Send a Revert command.
 *
 * This tells the Talker to do an RTI using passed stack address,
 * which starts it running with all registers set, until it decides
 * to stop.
 *
 * We detect that because it sends the STOPPED character (^C),
 * followed by the new stack value (low byte then high).
 */
boolean
TalkerRevert(ushort sp)
{
	uchar		revertcmd[3];

	revertcmd[0] = 0x38;
	if (!TalkerSendCheckEcho(revertcmd[0]))
	{
		printf("Revert command not properly acknowledged\n");
		return FALSE;
	}
	revertcmd[1] = sp>>8;
	revertcmd[2] = sp;
	mcu->write(revertcmd+1, 2);		// No echo is expected
	eaddr = sp;
	stopping_state = ST_Run;
	return TRUE;
}

/*
 * Set the talker running, returning after it stops.
 *
 * Normal console input and output are performed, including
 * tilde (~) escaped commands to db11, with the exception that
 * receipt of a STOPPED character (^C) is taken to mean that a
 * breakpoint was hit.
 *
 * REVISIT: a future version could require the two bytes that
 * follow STOPPED to come quickly, or else regard it as normal
 * program output and continue.  This timing-dependency might
 * be preferable to the alternative of falsely assuming that a
 * breakpoint was hit.
 */
boolean
TalkerRun()
{
	Multiplexed*	mpx_pp[2];
	boolean		done = FALSE;

	if (!TalkerRevert(stack))
		return FALSE;	// Failed

	while (!done)
	{
		mpx_pp[0] = stdio;
		mpx_pp[1] = mcu;

		if (!Multiplex(mpx_pp, 2))
			continue;

		/* Process all data received: */
		done = FALSE;

		if (mcu->isReady()	// Handle data received from talker
		 && !processTalkerInput())
		 	done = TRUE;

		if (stdio->isReady()	// Handle data received from keyboard
		 && !processKeyInput())
		 	done = TRUE;
	}
	return TRUE;
}
