/*
 * $Id: util.cxx,v 1.12 2002/03/17 11:08:08 cjh Exp $
 *
 * Utility functions for db11
 *
 * Copyright (c) 1996-2000, Clifford Heath.
 *
 * This program may be used for any commercial or noncommercial purpose
 * without charge, but acknowledgement of authorship and this copyright
 * message must be retained in any copies.
 */
#include	<stdincs.h>

#include	<util.h>
#include	<serial.h>

void
Fatal(char* fmt, ...)
{
	va_list	ap;

	printf("Fatal error: ");
	va_start(ap, fmt);
	vprintf(fmt, ap);
	printf("\n");
	exit(EXIT_FAILURE);
}

void
Notify(char* fmt, ...)
{
	va_list	ap;

	va_start(ap, fmt);
	vprintf(fmt, ap);
}

void
Reset(Serial* sio, char* fmt, ...)
{
	va_list	ap;

	va_start(ap, fmt);
	vprintf(fmt, ap);
	printf("\nRESET, then press Return to continue: ");
	while (getchar() != '\n')
		;
	sio->discard();		/* Flush all input and output. */
}

/*
 * Return a (static) printable string representing the character
 */
char*
Ascii(char ch, boolean passwhite)
{
	static char	buf[20];

	if (passwhite
	 && (!(isspace(ch) && ch != '\f') && (unsigned char)ch < ' '))
	{		/* Control character */
		sprintf(buf, "^%c", ch+'@');
	}
	else if (!isprint(ch) && !isspace(ch))
	{		/* Other non-printable character */
		sprintf(buf, "\\x%02X", ch&0xFF);
	}
	else		/* Normal printable character */
		sprintf(buf, "%c", ch);
	return buf;
}

/*
 * The C library provides a strdup() that uses malloc - we must use new!
 */
char*
strdup(const char* s)
{
	char*	copy = new char[strlen(s)+1];
	strcpy(copy, s);
	return copy;
}

char*
concat(const char* s1, const char* s2)
{
	char*	cp = new char[strlen(s1)+strlen(s2)+1];

	strcpy(cp, s1);
	strcat(cp, s2);
	return cp;
}

/*
 * Return offset of first differing byte in memory blocks s1, s2.
 */
int
memdiff(const char* s1, const char* s2, int max)
{
	int	i;

	for (i = 0; i < max; i++)
		if (*s1++ != *s2++)
			break;
	return i;
}

/*
 * Chew a line into an argv array of "words".
 * Handles the rules for single and double quoted strings,
 * including backslash escapes for enclosed quotes.
 * Non-quoted "magic" characters become words by themselves.
 *
 * Returns a pointer to a static argv array, which gets deleted on next call.
 */
char**
strcmd(
	const char*	line,
	const char*	magic,
	int*		argc_p
)
{
	const char*	cp = line;
	const char*	ep;
	char		quote;
	int		argc = 0;
	char**		argv;
	int		argalloc;

	// Allocate new argv array
	argv = new char*[argalloc = strlen(line)+1];
	memset(argv, 0, sizeof(char*)*argalloc);

	while (*cp != '\0')
	{		// Process entire string.
		while (isspace(*cp))
			cp++;		// Skip initial space
		if (*cp == '\0')
			break;		// Finished on white space

		ep = cp;		// Find end of the "word"
		if (strchr(magic, *ep))
		{		// "Magic character" word
			argv[argc] = new char[2];
			argv[argc][0] = *ep++;
			argv[argc][1] = '\0';
			argc++;
			cp = ep;
			continue;
		}

		while (*ep != '\0'		// Not end of line
		    && !isspace(*ep)		// or space
		    && strchr(magic, *ep) == 0	// or magic
		)
		{		// Pass one "word"
			if (*ep == '"' || *ep == '\'')
			{		// Pass a quoted string
				quote = *ep++;
				while (*ep != '\0' && *ep != quote)
				{
					if (*ep == '\\')
					{		// Allow escaped quote
						if (*++ep == '\0')
							break; // end of line!
					}
					ep++;
				}
				if (*ep != '\0')
					ep++;
			}
			else
				ep++;
		}

		/*
		 * Save a copy of the word; ep points to byte after
		 */
		argv[argc] = new char[ep-cp+1];
		strncpy(argv[argc], cp, ep-cp);
		argv[argc++][ep-cp] = '\0';

		if (*ep == '\0')
			break;
		cp = ep;
	}
	argv[argc] = (char*)0;
	if (argc_p)
		*argc_p = argc;
	return argv;
}

void
strcmdfree(char** argv)
{
	if (!argv)
		return;
	for (char** ap = argv; *ap; ap++)
		delete *ap;
	delete [] argv;
}

#ifdef	TEST_STRCMD
main()
{
	char	buf[256];
	while (printf(">"), gets(buf))
	{
		int	argc;
		char**	argv = strcmd(buf, "#", &argc);

		printf("%d words:\n", argc);
		while (*argv != (char*)0)
			printf("\t%s\n", *argv++);
	}
}
#endif

int
ReadS19(FILE* fp, S19* sp)
{
	char*	err;
	char	buf[81];
	char*	bp;
	int	checksum;
	int	result;

	sp->address = 0;
	sp->bytes = 0;

	if (!fgets(buf, sizeof(buf), fp))
		return 0;	/* EOF */

	if (*buf != 'S')
	{
		err = "not an S record";
		goto bad;
	}

	if (buf[1] == '0')
		return '0';	/* S0 records carry no data */

	if (sscanf(buf+2, "%2x", &checksum) != 1)
	{
		err = "can't read checksum";
		goto bad;
	}

	switch (buf[1])
	{
	case '1':
		result = sscanf(buf+4, "%4x", &sp->address);
		bp = buf+8;
		break;

	case '2':
		result = sscanf(buf+4, "%2x", &sp->address);
		bp = buf+6;
		break;

	case '3':
		result = sscanf(buf+4, "%8x", &sp->address);
		bp = buf+12;
		break;

	case '9':
		return 0;	/* End of file */

	default:
		err = "bad record type";
		goto bad;
	}

	if (result != 1)
	{
		err = "Can't read address field";
		goto bad;
	}

	for (sp->bytes = 0; sp->bytes < (int)sizeof(sp->data); sp->bytes++)
	{
		if (*bp == '\r' || *bp == '\n' || *bp == '\0'
		 || bp[1] == '\r' || bp[1] == '\n' || bp[1] == '\0')
			break;
		if (sscanf(bp, "%2x", &result) != 1)
		{
			err = "bad data";
			goto bad;
		}
		bp += 2;
		sp->data[sp->bytes] = result;
	}
	sp->bytes--;	/* last byte is extra */
	return buf[1];

bad:
	Fatal("Illegal S record: %s", err);
	return 0;
}

void
SubtractTimeval(struct timeval* from, struct timeval* subtract)
{
	if (from->tv_usec < subtract->tv_usec)
	{
		from->tv_usec += 1000000;
		from->tv_sec--;
	}
	from->tv_usec -= subtract->tv_usec;
	from->tv_sec -= subtract->tv_sec;
}

void
TraceTime(
	const char*
#if	defined(TRACING)
		fmt
#endif
	, ...
)
{
#if	defined(TRACING)
	static	FILE*		fp;
	static	struct	timeval	start;
	static	struct	timeval	last;
	struct	timeval		tv;
	struct	timeval		now;
	struct	timeval		delta;
	va_list			ap;

	if (!fp)
	{
		fp = fopen("/tmp/db11.trc", "w");
		(void)gettimeofday(&start, (struct timezone*)0);
		last = start;
	}

	(void)gettimeofday(&tv, (struct timezone*)0);
	now = tv;
	delta = tv;

	/*
	 * Find delta since start, and last:
	 */
	SubtractTimeval(&tv, &start);
	SubtractTimeval(&delta, &last);
	last = now;

	fprintf(
		fp, "%ld.%06ld (+%ld.%06ld) ",
		tv.tv_sec, tv.tv_usec,
		delta.tv_sec, delta.tv_usec
	);
	va_start(ap, fmt);
	vfprintf(fp, fmt, ap);
	va_end(ap);

	fflush(fp);
#endif
}

#if	defined(NEED_SNPRINTF)
int snprintf(char *str, size_t size, const  char  *format, ...)
{
	va_list	ap;
	int	i;

	va_start(ap, format);
	i = vsprintf(str, format, ap);
	va_end(ap);
	return i;
}
#endif
