;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; $Id: protocol.s,v 1.4 2002/03/17 11:08:09 cjh Exp $
;
; 68hc11 debug monitor (resident part)
;
; This program fits in the first 256 bytes of memory in a 68hc11, and can
; read&write memory, program internal ROM, run the program in any state,
; and resume control after a breakpoint.  External memory may be accessed
; (assuming that expanded mode is in effect), and external ROM programmed
; by memory reads and writes using any algorithm known to the host part.
;
; Copyright (c) 1997, Clifford Heath.
;
; This program may be used for any commercial or noncommercial purpose
; without charge, but acknowledgement of authorship and this copyright
; message must be retained in any copies.
;
; Commands and protocol:
;
; REVERT	0x38
;	Echo the command byte, then receive a two-byte stack pointer
;	value (MSB then LSB) and execute a Return From Interrupt (rti).
;
; SETVAR	0x39-0x3F
;	Echo the command byte, then receive one byte and store it in one
;	of the seven variable bytes. The variables are used in other
;	commands, and are:
;	EADDR	(2 bytes, set by command 0x39/0x3A)
;		Address of next byte to program, get or put.
;	EREG	(2 bytes, set by command 0x3B/0x3C)
;		Address of the programming control register (PCR)
;	EPREP	(1 byte, set by command 0x3D)
;		Value to store in PCR before initiating program operation.
;		For internal (E)EPROM, this has (E)ELAT set.
;	ESTART	(1 byte, set by command 0x3E)
;		Value to store in PCR to start programming (apply prog voltage).
;		For internal (E)EPROM, this has (E)ELAT and (E)EPROG set.
;	EDONE	(1 byte, set by command 0x3F)
;		Value to store in PCR to end programming. Often zero.
;	
; GET		0x40-0x7F
;	Echo the command byte, then send 1-64 bytes of data from the
;	address set in EADDR, incrementing EADDR. The low six bits of
;	the command give the requested byte count minus one.
;
; PUT		0x80-0xBF
;	Echo the command byte, then receive and store 1-64 bytes of data
;	from the address set in EADDR, incrementing EADDR. Each byte is
;	read back and transmitted. The low six bits of the command give the
;	requested byte count minus one.
;
; PROG		0xC0-0xFF
;	Echo the command byte, then receive and program 1-64 bytes of data
;	from the address set in EADDR, incrementing EADDR. The low six bits
;	of the command give the requested byte count minus one.
;
;	The programming algorithm for each byte is:
;		*EREG = EPREP;
;		receive a byte to program;
;		*EADDR = the received byte;
;		*EREG = ESTART;
;		receive input characters until a non-zero byte is received;
;		*EREG = EDONE;
;		re-read and transmit the value after programming;
;		EADDR = EADDR+1;
;
;	Note that the programming time is determined by the receive baud
;	rate, the number of idle (zero) characters received during
;	programming, and any inter-character delays.
;
; Software Interrupts:
;
; When the software interrupt handler is reached, the talker transmits the
; character code "STOPPING" (^C), followed by the stack pointer (low byte
; first!).  Additionally, the stack pointer is stored in EADDR, so assuming
; that there actually is an interrupt stack frame on the top of the stack,
; a "GET 8" (0x47) command will retrieve (one trash byte and) the stacked
; registers.
;
; A program may wish to use SCI interrupts, and/or detect an interrupt
; character from the host. It should ensure that the registers are stacked
; in a normal interrupt stack frame at the top of the stack, that the SCI
; interrupts are disabled (this can be by the CCR I-bit being set, a normal
; side-effect of an interrupt routine being called) then should jump to the
; software interrupt location in the talker. For a boot-RAM talker on a chip
; from series A-F, the SWI vector is available in the two bytes at address
; 0xF5, otherwise it is at 0xFFF6. This will cause the talker to send the
; STOPPING character and the SP, and then enter command mode. The user's
; program may be restarted by a Revert command from the host, using the
; stack pointer value that was received, and this will re-enable the
; interrupts normally.
;
; Startup:
;
; The program starts by waiting for a byte (any byte) from the host. It then
; acts as though an SWI occurred, sending "STOPPING" and the stack pointer,
; before entering the command loop.  In this way a host program can handle
; an alternate version that sets the stack elsewhere.
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Initialisation code, once only; 17 bytes
; For an (E)EPROM resident version, stack might be already set by "starter"
; In this case we won't "lds"
Initialise:
		lds	#STACK_START	; Set stack pointer
		clr	REGS+BPROT	; Allow EEPROM programming
; Setup SCI Transmitter and Receiver, to 9600, no interrupts
		ldd	#BAUDRATE*256+SCI_ENABLE
		staa	REGS+BAUD	; With 8MHz crystal, baud=9600.
		stab	REGS+SCCR2	; transmitter and receiver, no intrs
Synchronise:	jsr	*INSCI		; Wait until host speaks
		bne	Synchronise	; Not Null-byte, try again
;		jmp	SWI_Interrupt	; Announce our stack

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; SWI_Interrupt; Tell host program
SWI_Interrupt:
		ldab	#STOPPING	; Say we stopped
		jsr	*OUTSCI		; to the host program
		sts	*EADDR		; Store stack pointer
		ldd	*EADDR		; Get it back
		jsr	*OUTSCI		; Send low byte
		tab			; switch regs around
		jsr	*OUTSCI		; Send high byte
;		bra	WaitCommand	; Fall through

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Wait for a command, main command loop
WaitCommand:
		jsr	*INSCI		; Get command into B
		jsr	*OUTSCI		; Echo command
		ldx	*EADDR		; Get location to process
		addb	#0x40		; If this overflows...
		bcs	ProgRom		; it's a ProgramROM command
		addb	#0x40		; If this overflows...
		bcs	PutMem		; it's a Put Memory command
		addb	#0x40		; If this overflows...
		bcs	GetMem		; it's a Get Memory command
		addb	#VARS_SIZE	; If this overflows...
		bcs	SetVar		; it's a SetVar command
		incb			; if this doesn't overflow to zero
		bne	SWI_Interrupt	; it's a bad command, otherwise
;		bra	RevertCmd	; it's a Revert command
; Fall through... B is zero.

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Revert to program by setting stack pointer and returning as from an interrupt
RevertCmd:
		jsr	*INSCI		; Get high byte of new SP
		tba
		jsr	*INSCI		; Get low byte of new SP
		std	*EADDR		; Load the value into SP...
		lds	*EADDR
NULLSRV:				; Null interrupt handler
		rti			; Restart from stack frame

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Set Variable command handler - register A contains the variable number
SetVar:		ldx	#VARS_BASE
		abx			; Add B register
		jsr	*INSCI		; Get character
		jsr	*OUTSCI		; Echo the value stored
		stab	0,X		; Store B into variable
		bra	WaitCommand	; Continue

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; GetMem; B contains number of bytes to send minus one, X contains location
GetMem:		tba			; Counter in A
GetMemLoop:	ldab	0,x		; Read memory into B
		jsr	*OUTSCI
		inx			; Next location
		deca
		bge	GetMemLoop

; GetMem and PutMem come here when done:
StoreEADDR:	stx	*EADDR
		bra	WaitCommand

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; PutMem; B contains number of bytes to receive minus one, X contains location
PutMem:		tba			; Counter into A
PutMemLoop:	jsr	*INSCI		; Get byte
		stab	0,x		; Store value to memory
		deca			; Decrement count
		blt	StoreEADDR	; Done? Then save finishing address.
		ldab	0,x		; Check we can read it back
		inx			; Next location
		jsr	*OUTSCI		; Echo the character read back
		bra	PutMemLoop	; More to follow

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Get character from SCI into B register
INSCI:
		ldab	REGS+SCSR	; Wait for RDRF
;		bitb	#(FRAMINGERR+OVERRUN) ; if break detected
;		bne	USERSTART	; restart talker
		andb	#RDRF		; character available?
		beq	INSCI		; not yet
		ldab	REGS+SCDR	; get it
		rts

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Put a character to SCI from B register
OUTSCI:
		psha			; save A
OUTSCI_Loop:	ldaa	REGS+SCSR	; Wait for TDRE
		anda	#TDRE		; output ready?
		beq	OUTSCI_Loop	; not yet
		pula			; restore A
		stab	REGS+SCDR	; send it, *SETTING CONDITION CODES*
		rts
